#!/usr/bin/env python
# -*- coding: iso-8859-15 -*-
#
# Distributes jobs into host machines
# 
# Esa Pitknen 2005
# Markus Heinonen 1.6.2006
#                 objectified 2008
#
# Veli Makinen 1.2.2011
#                 assuming private/public key for ssh  
#                 added blocking mode
#                 using only necessary amount of hosts
#
# Johannes Ylinen 1.3.2012
#		  added per-task prejobs
############################################################


import pexpect, optparse, sys, os, time, getpass, time, re


#
# class to handle SSH-connections
# opens a connection to ssh-host and keeps it alive until closed
# allows for blocking and non-blocking commands to be sent
#
class SSH:
	DEVNULL = "> /dev/null 2>&1"
	PROMPT = "[$]"
	
	# produces something like:
	# "maxmem:3798,freemem:1555,cores:2,cpuload:0.46"
	MAXMEM = "echo -n 'maxmem:' && free -m | head -2 | tail -1 | tr -s ' ' | cut -d ' ' -f2"
	FREEMEM = "echo -n freemem: $(( ( ` cat /proc/meminfo | grep Inactive: | tr -s ' ' | cut -f 2 -d ' ' ` + ` cat /proc/meminfo | grep MemFree: | tr -s ' ' | cut -f 2 -d ' ' `) / 1000 ))"
	CORES = "echo -n 'cores:' && grep processor /proc/cpuinfo | wc | tr -s ' ' | cut -d ' ' -f2"
	CPULOAD = "echo -n 'cpuload:' && uptime | tr -s ' ' | tr -d ',' | cut -d ' ' -f12"
	COMMA = r"| tr -d '\n' && echo -n ','"
	STATUSCMD = r"%s%s&&%s%s&&%s%s&&%s" % (MAXMEM, COMMA, FREEMEM, COMMA, CORES, COMMA, CPULOAD)
	
	EXIT_CMD = "exit"
	
	CONNECTION_OK = 1
	NO_CONNECTION = 2
	CONNECTION_HANG = 4
	CONNECTION_ERROR = 8
	PASSWORD_DENIED = 16
	NO_FS = 32
	RESERVED = 64
	
	# opens a process to 'host'
	def __init__(self, user, password, hostname, timeout=2.0):
		self.user = user
		self.password = password
		self.hostname = hostname
		self.timeout = timeout
		self.buffer = None
		self.connection = False
		self.statuscode = None
		self.donebuffer = []
		
		# open ssh process
		self.buffer = pexpect.spawn("ssh %s@%s" % (self.user, self.hostname))
		
		# either fails, asks for password or wants acknowledgement of host id
		#k = self.buffer.expect(['[Pp]assword:', "yes/no", pexpect.TIMEOUT, pexpect.EOF], timeout=3.0)
		
		#if k == 1: # answer 'yes' and wait for password prompt
		#	# assume no problems can occur here			 
		#	self.buffer.sendline("yes")
		#	self.buffer.expect("[Pp]assword:", timeout=3.0)
		#elif k == 2: # timeout -> abort connection attempt
		#	self.buffer.terminate()
		#	self.buffer.close()
		#	self.status = SSH.NO_CONNECTION
		#	return
		#elif k == 3: # end-of-file -> something went wrong and process quitted -> abort
		#	self.buffer.terminate()
		#	self.buffer.close()
		#	self.status = SSH.CONNECTION_ERROR
		#	return
		
							
		# send password
		#self.buffer.sendline(password)
		
		# wait for prompt
		
                # using public key / private key with ssh, no passwords submitted
		k = self.buffer.expect([SSH.PROMPT, "Password:", pexpect.TIMEOUT, pexpect.EOF], timeout=30.0)
		
		
		if k == 0: # got prompt, everything ok
			self.connection = True
			self.status = SSH.CONNECTION_OK
		elif k == 1 or k == 2 or k == 3: # timeout, assume crashed process
			self.buffer.terminate()
			self.buffer.close()
			self.status = SSH.PASSWORD_DENIED
		
		# now hopefully connection is open
		# also check for additional messages in the buffer (i.e. fs-error)
		
		# no fs connection
		if "chdir" in self.buffer.before:
			self.status = SSH.NO_FS
			self.close()
			return
		
		# reserved for some other use
		if "reserved" in self.buffer.before:
			self.status = SSH.RESERVED
			self.close()
			return None
		
		if "error" in self.buffer.before:
			self.status = SSH.CONNECTION_ERROR
			self.close()
			return
	
	
	def getstatus(self):
		return self.statuscode
	
	# gives the number of cores, cpu-load and free mem
	def gethostload(self):
		try:
			output = self.command_blocking(SSH.STATUSCMD)
			line = ""
			for l in output:
				if "cores:" in l:
					line = l
					break
		
			fields = dict( [(entry.split(":")[0],entry.split(":")[1]) for entry in line.strip().split(",")] )
		
			return (int(fields["cores"]), float(fields["cpuload"]), int(fields["maxmem"]), int(fields["freemem"]))
		except:
			# There was an error, report full load
			return (1, 1.00, 1, 0)
	
	
	def alive(self):
		# whether connection is opened and is still alive?
		return self.connection and self.buffer.isalive()
	
	# polls for output from the processes
	# if process is complete, we should see END-MARKER and prompt in output
	def poll(self):
		if self.connection:
			self.buffer.sendline()
			time.sleep(0.3) # wait for 200ms
			
			try:
				data = self.buffer.read_nonblocking(size=100000, timeout=0)
				data = data.strip().replace("\r","")
				
				if data == "":
					return None
				
				return data
			except:
				# timeout error, need to poll again
				return None
		
		return None
	
	
	# redirects to blocking command
	def command(self, cmd):
		return self.command_blocking(cmd)
	
	# executes a command, doesn't wait for results, directs output to /dev/null
	# check with "poll()" whether any processed have finished
	def command_nonblocking(self, cmd, cmd_id):
		# direct to /dev/null
		cmd = "/bin/bash -c \"%s\"&" % (cmd)
		
		# start in background
		cmdlen = self.buffer.sendline(cmd)
		
		# read until we see the process number indicator from background task
		# also match everything afterwards
		# -> .before contains all done's before our [pid]
		#    .after contains our [pid] and also all done's after pid
		k = self.buffer.expect([r"\[[0-9]+\][ ]+[0-9]+.*", pexpect.EOF], timeout=None)
		
		# success!
		if k == 0:
			# previous commands might have completed
			# -> check for '[xx] Done' in the buffer
			# -> append them to 'self.donebuffer'
			
			b = self.buffer.before
			a = self.buffer.after
			# get pid
			pid = int( a[a.find("[")+1:a.find("]")] )
			
			# get doneids 
			doneids = map(int, re.findall(r"\[([0-9]+)\][+-]?\s+Done" , b+a))
			self.donebuffer.extend(doneids)
			
			return pid
		# ssh connection crashed
		elif k == 1:
			self.connection = False
			return None
		
	
	# blocking execution of a command, waits for results
	# captures all output
	def command_blocking(self, cmd, cmd_id = 0):
		START_MARKER = "--START%d--" % (cmd_id)
		END_MARKER = "--END%d--" % (cmd_id)
		START_CMD = "echo %s" % (START_MARKER)
		END_CMD = "echo %s" % (END_MARKER)
		
		cmd = "/bin/bash -c \"%s && %s && %s\"" % (START_CMD, cmd, END_CMD)
		
#		print cmd
		
		l = self.buffer.sendline(cmd)
		buf1 = self.buffer.read(l)
		buf2 = ""
		buf = ""
		
		while START_MARKER not in buf: # if end marker isn't present
			c = self.buffer.read(1) # blocking, but doesn't matter
			if c not in ["\n", "\r", " "]: # for some reason, sendline adds extra \r's to the stream
				buf += c
		
		# now the stream only contains direct result from our command (or some spaces first)
		# -> read until first END_MARKER is echoed
		k = self.buffer.expect([END_MARKER, SSH.PROMPT, pexpect.EOF], timeout=None)
		
		if k == 0:
			buf2 = self.buffer.before
		elif k == 1:
			buf2 = self.buffer.before
		elif k == 2:
			self.close()
			return None
		
		# find if any "Done"'s are found
		buf = buf1.replace("\r", "").replace("\n","").replace(" ", "") + buf2.replace("\r","").replace("\n","").replace(" ", "")
		doneids = map(int, re.findall(r"\[([0-9]+)\][+-]?Done", buf))
		
		if "[" in buf:
			print buf
		
		if doneids:
			print "found doneids during update", doneids
			print "BUF", buf
		
		self.donebuffer.extend(doneids)
		
		# parse and return output
		buf2 = buf2.replace("\r", "").strip()
		lines = map(str.strip, buf2.split("\n"))
		
		output = []
		for l in lines:
			if not re.search(r"\[([0-9]+)\][+-]?\s+Done", l):
				output.append(l)
		
		return output
	
	
	def close(self):
		self.buffer.sendline(SSH.EXIT_CMD)
		self.buffer.terminate()
		self.buffer.close()
		self.connection = False




#
# Host-class
#

class Host:
	idcounter = 1
	
	def __init__(self, name):
		self.id = Host.idcounter
		Host.idcounter += 1
		self.name = name
		self.ssh = None
		self.activejobs = []
		self.corecount = None
		self.cpuload = None
		self.freecpu = None
		self.maxmem = None
		self.freemem = None
	
	
	def open(self, user, password):
		self.ssh = SSH(user, password, self.name)
		
		if self.ssh.connection == False:
			return
	
	
	# check whether ssh connection is alive
	def alive(self):
		if not self.ssh:
			return False
		
		return self.ssh.alive()	
	
	
	# update cpuload, freecpu, freemem 
	def update_status(self):
		if self.alive():
			self.corecount, self.cpuload, self.maxmem, self.freemem = self.ssh.gethostload()
			self.freecpu = self.corecount - self.cpuload
	
	
	# returns computation slots
	def slots(self, cpureq=1.0, memreq=1024):
		maxslots = self.maxmem / ( (len(self.activejobs) + 1) * memreq )
		return max([0, min([int(self.freecpu/cpureq), int(self.freemem/memreq), int(maxslots)])])
	
	
	# starts a job
	def start_job(self, j, blocking=True):
		os.system("ssh " + self.name + " \"" + j.precmd +"\"")
		if blocking:
			try:
				self.ssh.command_blocking(j.cmd)
			except:
				return None
			j.done = True
		else:
			pid = self.ssh.command_nonblocking(j.cmd, j.id)
			if pid is None: # failed somehow
				return
			
			self.activejobs.append(j)
			j.assigned = True
			j.done = False
			j.processid = pid
	
	
	# returns a list of finished jobs
	# updates the 'activejobs' list
	def check_dones(self):
		donejobs = []
		results = self.ssh.poll()
		
		# first check if stdout has got any new "Done"s
		if results:
			re.findall("\[([0-9]+)\] [0-9]+", results)
			doneids = map(int, re.findall(r"\[([0-9]+)\][+-]?\s+Done", results))
			
			for j in self.activejobs:
				if j.processid in doneids:
					donejobs.append(j)
		
		for j in self.activejobs:
			if j.processid in self.ssh.donebuffer:
				donejobs.append(j)
		
		self.ssh.donebuffer = [] # empty donebuffer
		
		for j in donejobs:
			j.endtime = time.time()
			j.done = True
			j.assigned = False
			j.processid = None
			if self.activejobs.count(j) > 0:
				self.activejobs.remove(j)
		
		return donejobs
	
	
	
	def close(self):
		self.ssh.close()
		self.ssh = None
	
	def status_str(self):
		if self.alive():
			return "%s\tavailable\tcpu load: %.2f/%d, mem available: %dmb" % (self.name, self.cpuload, self.corecount, self.freemem)
		else:
			return "%s\tnot available" % (self.name)
	
	def __str__(self):
		return self.name



#
# Job-class
#
# has essentially three states: 
# - not assigned
# - running
# - done 
# there's no way to detect when the process has crashed
#
class Job:
	idcounter = 1
	
	def __init__(self, jobstr, prejobstr):
		self.cmd = jobstr
		self.precmd = prejobstr
		self.starttime = None
		self.endtime = None
		self.assigned = False
		self.done = False
		self.processid = None
		self.id = Job.idcounter
		Job.idcounter += 1
	
	def __str__(self):
		return self.cmd

#
# Log class
#
class Log:
	def __init__(self, logfile, verbose=False):
		if logfile:
			self.logfile = open(logfile, "a")
		else:
			self.logfile = open(os.devnull, "w")
		self.verbose = verbose
	
	def write(self, s):
		if self.verbose:
			print s
		self.logfile.write(s + "\n")
	
	def __del__(self):
		self.logfile.close()


def find_next_job(joblist):
	for j in joblist:
		if not j.done and not j.assigned:
			return j
	return None





#
# variables
#

version = "1.2"
usage = "usage: ./%prog [-lmcuvqbps] HOSTFILE [JOBFILE]"
description = "Distributes jobs across host machines: SSH's into all host machines, starts the jobs and waits for job processes to finish; requeues crashed jobs. Write or redirect computation results directly into files, stdout is redirected to /dev/null"

#
# arguments
#

parser = optparse.OptionParser(usage=usage,version=version,description=description)
parser.add_option("-o", "--hostfile", dest="hostfile", type="string", help="File containing a list of host machines addresses (one per row)")
parser.add_option("-j", "--jobfile", dest="jobfile", type="string", help="File containing a list of jobs (scripts, executables, etc)")
parser.add_option("-l", "--logfile", dest="logfile", type="string", default="", help="Use log file")
parser.add_option("-m", "--memreq", dest="memreq", type="int", default=1024, help="Memory requirement of a single task [default %defaultMb]")
parser.add_option("-c", "--cpureq", dest="cpureq", type="float", default=0.95, help="Cpu requirement of a single task [default %default]")
parser.add_option("-u", "--user", dest="user", type="string", default=getpass.getuser(), help="Username for SSH [default '%default']")
parser.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False, help="Verbose mode")
parser.add_option("-q", "--quiet", dest="quiet", action="store_true", default=False, help="Quiet mode")
parser.add_option("-b", "--blocking", dest="blocking", type="int", default=0, help="Blocking mode")
parser.add_option("-p", "--prejob", dest="prejob", type="string", default="", help="Command file doing prejob for each host")
parser.add_option("-w", "--varyingprejob", dest="varyingprejobfile", type="string", default="", help="Path to per-task prejobs")
parser.add_option("-s", "--postjob", dest="postjob", type="string", default="", help="Command file doing postjob for each host")

(options, args) = parser.parse_args()


if options.hostfile == None:
	parser.error("No host file given")
if options.jobfile == None:
	parser.error("No job file given")
if len(args) > 0:
	parser.error("Unspecified arguments given")

if not os.path.isfile(options.hostfile):
	parser.error("hostfile '%s' not found" % (options.hostfile))
if not os.path.isfile(options.jobfile):
	parser.error("jobfile '%s' not found" % (options.jobfile))
if options.varyingprejobfile != None and not os.path.isfile(options.varyingprejobfile):
	parser.error("Per-task prejobfile '%s' not found" % (options.varyingprejobfile))

# dev-nullify output
if options.quiet:
	sys.stdout = open(os.devnull, "w")

password = ""

# assuming public key / private key usage with ssh
#getpass.getpass("SSH password for user '%s':" % (options.user))

#
# log file
#

logfile = Log(options.logfile, options.verbose)

#
# parse jobs
#

jobs = []
jfile = open(options.jobfile)
vpjfile = open(options.varyingprejobfile)
while True:
	jline = jfile.readline()
	vpjline = vpjfile.readline()
	if len(jline) == 0:
		break
	if jline.startswith("#") or jline == "" or vpjline.startswith("#") or vpjline == "":
		continue
		
	# no redirection -> append "> /dev/null"
	if ">" not in jline:
		jline += " > /dev/null 2>&1"
	if ">" not in vpjline:
		vpjline += " > /dev/null 2>&1"

	jobs.append( Job(jline, vpjline) )
jfile.close()
vpjfile.close()

#
# parse hosts
#

hosts = []
hfile = open(options.hostfile)
for line in hfile:
	line = line.strip()
	if line.startswith("#") or line == "":
		continue
	hosts.append( Host(line) )
hfile.close()

print "%d jobs and %d hosts" % (len(jobs), len(hosts))

print "Assuming %dMb memory and %.2f cpu requirements for jobs" % (options.memreq, options.cpureq)

#
# check hosts for availability
#

print "Checking hosts status..."

validhosts = 0
totalslots = 0
lasthost = 0
for h in hosts:
           h.open(options.user, password) # starts the control-ssh connections
	
	   if not h.alive():
		if h.ssh.status == SSH.PASSWORD_DENIED:
			print "Host %s\terror: password denied" % (h)
		elif h.ssh.status == SSH.CONNECTION_ERROR:
			print "Host %s\terror: connection crashed" % (h)
		elif h.ssh.status == SSH.NO_CONNECTION:
			print "Host %s\terror: no connection to host" % (h)
		elif h.ssh.status == SSH.NO_FS:
			print "Host %s\terror: no fs access" % (h)
		else:
			print "Host %s\terror: unkwnown" % (h)
		
		h.ssh = None
		continue
	
	   validhosts += 1
	
	   h.update_status() # checks cpu/mem status
	
	   slots = h.slots(options.cpureq, options.memreq)
	   totalslots += slots
	   	
  	   print "Host %s\tavailable with %s slots (%5.2f/%2d free cpu and %d/%dMb mem)" % (h, str(slots).rjust(2), h.freecpu, h.corecount, h.freemem, h.maxmem)
	   
	   if totalslots > len(jobs):
	      enoughhosts = True
	      lasthost = hosts.index(h)
	      break

if lasthost>0:
   del hosts[lasthost+1:]
   
print "%d/%d hosts available with %d computation slots (%d iterations for %d jobs)" % (validhosts, len(hosts), totalslots, max([1,len(jobs)/validhosts]), len(jobs))

#
# assigning first round of jobs
#

print "Assigning jobs..."

jobs_to_assign = len(jobs)
jobs_done = 0
for h in hosts:
	if not h.ssh:
		continue
	
	if jobs_to_assign == 0:
		break
	
	print "Host %s\t" % (h),
	
	# ssh connection has crashed
	if not h.alive():
		print "SSH connection has been terminated"
		h.ssh = None
		continue
	
	h.update_status()
	slots = h.slots(options.cpureq, options.memreq)
	
	if options.prejob != "":
	   os.system(options.prejob + " " + h.name)

	print "Assigning %d jobs (status: %.2f/%d, %dMb)" % (min([slots,jobs_to_assign]), h.freecpu, h.corecount, h.freemem)
	while slots > 0 and jobs_to_assign > 0:
		j = find_next_job(jobs)
		if not j:
			jobs_to_assign = 0
			break
		
		print " Assigning job %d/%d: %s" % (j.id, len(jobs), j.cmd),
		slots -= 1
		jobs_to_assign -= 1
		h.start_job(j, blocking=options.blocking)
		
		if not options.blocking:
		   if j.assigned:
			print "[pid %d]" % (j.processid)
		   else:
			print "... failed"
		else:
		   if j.done:
		      print "... done"
		      jobs_done += 1
		   else:
		      print "... failed"

#
# start going through hosts:
# - poll whether any jobs finished
# - if available slots -> start more jobs
#

time.sleep(5)

print "Polling for results..."

hptr = 0
try:
  while jobs_done < len(jobs):
	# take next host
	h = hosts[hptr]
	hptr += 1
	if hptr == len(hosts):
		hptr = 0
	
	if not h.ssh:
		continue
	
	print "[%d/%d done]. Host %s\t" % (jobs_done, len(jobs), h),
	
	# check host status
	if not h.alive():
		# crashed, requeue all jobs
		h.ssh = None
		ids = []
		for j in h.activejobs:
			j.assigned = False
			j.done = False
			j.starttime = None
			j.endtime = None
			ids.append(j.id)
			jobs_to_assign += 1
		
		print "crashed, retrieving %d jobs [%s]" % (len(h.activejobs), ",".join(map(str,ids)))
		continue
	
	# otherwise host is alive
	# -> check if any jobs have finished
	
	donejobs = h.check_dones() # updates 'h.activejobs'
	if donejobs:
		jobs_done += len(donejobs)
		print "%d jobs done [%s]" % (len(donejobs), ",".join( [str(j.id) for j in donejobs] ))
	else:
		print "no finished jobs"
	
	# -> check for more computation slots
	
	if jobs_to_assign > 0:
		h.update_status()
		slots = h.slots(options.cpureq, options.memreq)
		
		print "[%d/%d done]. Host %s\tAssigning %d jobs (status: %.2f/%d, %dMb)" % (jobs_done, len(jobs), h, min([slots,jobs_to_assign]), h.freecpu, h.corecount, h.freemem)
		while slots > 0 and jobs_to_assign > 0:
			j = find_next_job(jobs)
			if not j:
				jobs_to_assign = 0
				break
			
			print " Assigning job %d/%d: %s" % (j.id, len(jobs), j.cmd),
			slots -= 1
			jobs_to_assign -= 1
			h.start_job(j, blocking=options.blocking)
			
   	 	        if not options.blocking:
		           if j.assigned:
		 	      print "[pid %d]" % (j.processid)
		           else:
			      print "... failed"
		        else:
		           if j.done:
		              print "... done"
		              jobs_done += 1
		           else:
		              print "... failed"
		
	# sleep 1 second until next polling
	time.sleep(1.0)

except:
   print "Interrupted: remember to kill all hanging processes."
   print "[%d/%d done]" % (jobs_done, len(jobs))
   print "Terminating SSH connections to %d hosts" % (validhosts)
   for h in hosts:
	if h.ssh:
           if options.postjob != "":
   	      os.system(options.postjob + " " + h.name)
           h.close()

   print "done"
   print "[%d/%d done]" % (jobs_done, len(jobs))
   raise

#
# Done
#

print "Terminating SSH connections to %d hosts" % (validhosts)

for h in hosts:
	if h.ssh:
           if options.postjob != "":
   	      os.system(options.postjob + " " + h.name)
           h.close()

print "done"
