// blasttablereporter.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//
// Additions / changes by Peter Husemann
//


#include "blasttablereporter.hpp"
#include "karlinaltschul.hpp"
#include "sequencedatabase.hpp"
#include "unique_keep.hpp"
#include "version.hpp"

#include <algorithm>
#include <cstdio>
#include <string>

namespace swift
  {

  namespace
    {
    using std::sort;
    using std::string;


    string const VersionLine =
      string("# SWIFT ") + Version::svnRevision() + "\n";
    string const FieldsLine =
      "# Fields:\n"
      "#Query ID\t Subject ID\t % Identity\t Alignment Length\t "
      "Mismatches\t Gap Openings\t Query Start\t Query End\t Subject Start\t "
      "Subject End\t E-Value\t Bit Score\n";

    string const FieldsLineCompAssemb =
      "#Query ID\t Query Length\t Subject ID\t Subject Length\t % Identity\t Alignment Length\t "
      "Mismatches\t Gap Openings\t Query Start\t Query End\t Subject Start\t "
      "Subject End\t E-Value\t Bit Score\n";


    /*!
        Functor for ordering of alignments.
    */
    struct AlignmentOrder
      {
public:
        bool operator()(
          BlastTableReporter::SubjectAlignment const & a,
          BlastTableReporter::SubjectAlignment const & b
        ) const;
      };


    inline
    bool AlignmentOrder::operator()(
      BlastTableReporter::SubjectAlignment const & a,
      BlastTableReporter::SubjectAlignment const & b
    ) const
      {
        if (a.alignment->score != b.alignment->score)
          return (a.alignment->score > b.alignment->score);
        else if (a.subjectId != b.subjectId)
          return (a.subjectId < b.subjectId);
        else if (a.alignment->queryBegin != b.alignment->queryBegin)
          return (a.alignment->queryBegin < b.alignment->queryBegin);
        else if (a.alignment->subjectBegin != b.alignment->subjectBegin)
          return (a.alignment->subjectBegin < b.alignment->subjectBegin);
        else if (a.alignment->queryEnd != b.alignment->queryEnd)
          return (a.alignment->queryEnd < b.alignment->queryEnd);
        else if (a.alignment->subjectEnd != b.alignment->subjectEnd)
          return (a.alignment->subjectEnd < b.alignment->subjectEnd);
        else
          return false;
      }


    /*!
        Functor for determining alignment equality.
    */
    struct AlignmentEquality
      {
public:
        bool operator()(
          BlastTableReporter::SubjectAlignment const & a,
          BlastTableReporter::SubjectAlignment const & b
        ) const;
      };


    inline
    bool AlignmentEquality::operator()(
      BlastTableReporter::SubjectAlignment const & a,
      BlastTableReporter::SubjectAlignment const & b
    ) const
      {
        // Note:
        // - queryId is not considered as we check within one query.
        // - bit-wise struct comparison not possible due to editScript.
        if ((a.subjectId               == b.subjectId) &&
            (a.alignment->queryBegin   == b.alignment->queryBegin) &&
            (a.alignment->subjectBegin == b.alignment->subjectBegin) &&
            (a.alignment->queryEnd     == b.alignment->queryEnd) &&
            (a.alignment->subjectEnd   == b.alignment->subjectEnd) &&
            (a.alignment->score        == b.alignment->score))
          return true;

        return false;
      }

  } // anonymous namespace


  /*!
      Constructs a BlastTableReporter object.

      \arg \c outStream Report output stream.
      \arg \c subjects Subject sequences.
      \arg \c queries Query sequences.
      \arg \c kaStats Karlin-Altschul statistics.
  */
  BlastTableReporter::BlastTableReporter(
    boost::shared_ptr<SequenceDatabase const> queries,
    boost::shared_ptr<SequenceDatabase const> subjects,
    boost::shared_ptr<KarlinAltschul const> kaStats,
    std::ostream & outStream
  ) :
      queries_(queries),
      subjects_(subjects),
      kaStats_(kaStats),
      outStream_(outStream)
{}


  /*!
      Destroys the BlastTableReporter object.
  */
  BlastTableReporter::~BlastTableReporter()
  {
    outStream_.flush();
  }


  void BlastTableReporter::clear()
  {
    for (AIterator i(alignments_.begin()), j(alignments_.end()); i != j; ++i)
      delete (*i).alignment;

    alignments_.clear();
  }


  void BlastTableReporter::append(
    Alignment * alignment,
    unsigned int queryId,
    unsigned int subjectId
  )
  {
    queryId_ = queryId;
    alignments_.push_back(SubjectAlignment(alignment, subjectId));
  }


  /*!
      Removes duplicate alignments.
  */
  void BlastTableReporter::removeDuplicates()
  {
    AIterator newEnd(
      unique_keep(
        alignments_.begin(),
        alignments_.end(),
        AlignmentEquality()
      )
    );

    for (AIterator i(newEnd), j(alignments_.end()); i != j; ++i)
      delete (*i).alignment;

    alignments_.erase(newEnd, alignments_.end());
  }


  void BlastTableReporter::report( bool &compassemb)
  {
    static unsigned int const Base(1); // BLAST positions are base-1
    static char eValueBuf[16];
    static char bitScoreBuf[16];

    if (isEmpty())
      return;

    sort(alignments_.begin(), alignments_.end(), AlignmentOrder());
    removeDuplicates();

    string qHeader(queries_->text(queryId_));
    string qId(qHeader.substr(0, qHeader.find_first_of(' ')));
    unsigned int qSize(queries_->size(queryId_));
    unsigned int sSize(subjects_->size());

    if (compassemb)
      { // different output format for comperative assembly project
        outStream_ << FieldsLineCompAssemb;
      }
    else
      {
        outStream_
        << VersionLine
        << "# Query: " << qHeader << "\n"
        << FieldsLine;
      }

    for (AIterator i(alignments_.begin()), j(alignments_.end()); i != j; ++i)
      {
        SubjectAlignment & sa(*i);
        Alignment * a(sa.alignment);

        string sHeader(subjects_->text(sa.subjectId));
        string sId(sHeader.substr(0, sHeader.find_first_of(' ')));

        formatEValue(kaStats_->eValue(a->score, qSize, sSize), eValueBuf);
        formatBitScore(kaStats_->bitScore(a->score), bitScoreBuf);

        outStream_
        << qId; // Query ID

        if (compassemb)
          { // different output format for comperative assembly project
            outStream_ << "\t"
            << qSize; // Query Size
          }

        outStream_ << "\t"
        << sId; // Subject ID

        if (compassemb)
          { // different output format for comperative assembly project
            outStream_ << "\t"
            << sSize; // Subject Size
          }

        outStream_ << "\tn/a\t" // % identity -- TODO: not implemented yet
        << a->subjectEnd - a->subjectBegin + 1 // Alignment Length
        << "\tn/a"    // mismatches -- TODO: not implemented yet
        << "\tn/a"    // gap openings -- TODO: not implemented yet
        << "\t"
        << a->queryBegin + Base // Query Start
        << "\t"
        << a->queryEnd + Base // Query End
        << "\t"
        << a->subjectBegin + Base // Subject Start
        << "\t"
        << a->subjectEnd + Base // Subject End
        << "\t"
        << eValueBuf // E-Value
        << "\t"
        << bitScoreBuf // Bit Score
        << "\n";

        // Print out alignment
        //         operator<<((std::ostream&)outStream_, (Alignment &) *a);
        //         outStream_ << "\n";

        // testing
        //         a->editScript->toString(outStream_);

        delete a;
      }

    outStream_ << "\n";

    alignments_.clear();
  }


  /*!
      Formats an E-value like BLAST.

      \arg \c eValue E-value to format.
      \arg \c buffer Buffer to write formatted E-value into.
  */
  void BlastTableReporter::formatEValue(double eValue, char * buffer) const
    {
      if (eValue < 1.0e-180)
        {
          sprintf(buffer, "0.0");
        }
      else if (eValue < 1.0e-99)
        {
          sprintf(buffer, "%2.0le", eValue);
        }
      else if (eValue < 0.0009)
        {
          sprintf(buffer, "%3.0le", eValue);
        }
      else if (eValue < 0.1)
        {
          sprintf(buffer, "%4.3lf", eValue);
        }
      else if (eValue < 1.0)
        {
          sprintf(buffer, "%3.2lf", eValue);
        }
      else if (eValue < 10.0)
        {
          sprintf(buffer, "%2.1lf", eValue);
        }
      else
        {
          sprintf(buffer, "%5.0lf", eValue);
        }
    }


  /*!
      Formats a bit score like BLAST.

      \arg \c bitScore Bit score to format.
      \arg \c buffer Buffer to write formatted bit score into.
  */
  void BlastTableReporter::formatBitScore(double bitScore, char * buffer) const
    {
      if (bitScore > 9999)
        {
          sprintf(buffer, "%4.3le", bitScore);
        }
      else if (bitScore > 99.9)
        {
          sprintf(buffer, "%4.0ld", (long)bitScore);
        }
      else
        {
          sprintf(buffer, "%4.1lf", bitScore);
        }
    }

      /*!
      Appends a message to the output.
       */
    void BlastTableReporter::message(const  char * m) const
    {
      outStream_ << m;
    }

} // namespace swift


