// fastastreamiterator.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#include "fastastreamiterator.hpp"
#include "streamutil.hpp"
#include <boost/algorithm/string/trim.hpp>
#include <cctype>
#include <string>

namespace swift {

namespace detail {

// FastaStreamIteratorImpl class declaration --------------------------------

struct FastaStreamIteratorImpl {
    
    FastaStreamIteratorImpl(std::istream * stream);
    ~FastaStreamIteratorImpl();
    void advance();
    bool next();
    bool nextDefline();
    bool nextSequence();
    FastaStreamIterator::const_reference reference();
    FastaStreamIterator::const_pointer pointer(); 
    bool equal(FastaStreamIteratorImpl * rhs);

    std::istream * stream_;
    char * rdBuffer_;
    FastaSequence fasta_;

    static size_t const RdBufferSize;
    static char const DeflineTag;
};

size_t const FastaStreamIteratorImpl::RdBufferSize = 16384;
char const FastaStreamIteratorImpl::DeflineTag = '>';

// FastaStreamIteratorImpl class definition---------------------------------

FastaStreamIteratorImpl::FastaStreamIteratorImpl(std::istream * stream) :
    stream_(stream),
    rdBuffer_(new char[RdBufferSize])
{
    advance();
}

// --------------------------------------------------------------------------

FastaStreamIteratorImpl::~FastaStreamIteratorImpl()
{
    delete [] rdBuffer_;
}

// --------------------------------------------------------------------------

void FastaStreamIteratorImpl::advance()
{
    if (stream_ && !next())
        stream_ = 0;
}

// --------------------------------------------------------------------------

bool FastaStreamIteratorImpl::next()
{
    return nextDefline() && nextSequence();
}

// --------------------------------------------------------------------------

bool FastaStreamIteratorImpl::nextDefline()
{
    bool tagFound(false);
    std::streamsize charsRead(0);
    std::streamsize i(0);
    std::string & defline(fasta_.defline);

    // Locate FASTA defline tag
    while (true) {
        charsRead = swift::readsome(*stream_, rdBuffer_, RdBufferSize);

        if (charsRead <= 0) {
            break;
        } else {
            i = 0;
            while (i < charsRead && ::isspace(rdBuffer_[i]))
                ++i;

            if (i < charsRead) {
                if (rdBuffer_[i] == DeflineTag) {
                    tagFound = true;
                    ++i;
                }
                break;
            }
        }
    }

    if (!tagFound)
        return false;

    // Read and trim defline
    defline.erase();
    std::streamsize j(0);
    while (true) {
        j = i;
        while (j < charsRead && rdBuffer_[j] != '\n' && rdBuffer_[j] != '\r')
            ++j;

        defline.append(rdBuffer_ + i, j - i);

        if (j == charsRead) {
            charsRead = swift::readsome(*stream_, rdBuffer_, RdBufferSize);
            i = 0;
        } else {
            swift::pushback(*stream_, rdBuffer_ + j, charsRead - j);
            break;
        }
    }

    boost::trim(defline);
    return defline.size();
}

// --------------------------------------------------------------------------

bool FastaStreamIteratorImpl::nextSequence()
{
    std::streamsize charsRead(0);

    std::string & seq(fasta_.sequence);
    seq.erase();

    while (true) {
        charsRead = swift::readsome(*stream_, rdBuffer_, RdBufferSize);

        if (charsRead <= 0) {
            break;
        } else {
            std::streamsize i(0);
            std::streamsize j(0);

            while (i < charsRead && rdBuffer_[i] != DeflineTag) {
                j = i;

                while (j < charsRead && !::isspace(rdBuffer_[j]))
                    ++j;

                size_t len(j - i);
                if (seq.capacity() < seq.size() + len) {
                    seq.reserve(2 * std::max(seq.capacity(), len));
                }
                seq.append(rdBuffer_ + i, len);
                
                i = j + 1;
                while (i < charsRead && ::isspace(rdBuffer_[i]))
                    ++i;
            }
        
            if (i < charsRead && rdBuffer_[i] == DeflineTag) {
                swift::pushback(*stream_, rdBuffer_ + i, charsRead - i);
                break;
            }
		}
	}

    return seq.size();
}

// --------------------------------------------------------------------------

FastaStreamIterator::const_reference FastaStreamIteratorImpl::reference()
{
    return fasta_;
}

// --------------------------------------------------------------------------

FastaStreamIterator::const_pointer FastaStreamIteratorImpl::pointer()
{
    return &fasta_;
}

// --------------------------------------------------------------------------

bool FastaStreamIteratorImpl::equal(FastaStreamIteratorImpl * rhs)
{
    return (rhs && stream_ == rhs->stream_);
}

} // namespace detail

// FastaStreamIterator class definition -------------------------------------

FastaStreamIterator::FastaStreamIterator() :
    pimpl_(boost::shared_ptr<ImplType>(new ImplType(0)))
{
}

// --------------------------------------------------------------------------

FastaStreamIterator::FastaStreamIterator(std::istream & inputStream) :
    pimpl_(boost::shared_ptr<ImplType>(new ImplType(&inputStream)))
{
}

// --------------------------------------------------------------------------

FastaStreamIterator::const_reference FastaStreamIterator::operator*() const
{
    return pimpl_->reference();
}

// --------------------------------------------------------------------------

FastaStreamIterator::const_pointer FastaStreamIterator::operator->() const
{
    return pimpl_->pointer();
}

// --------------------------------------------------------------------------

FastaStreamIterator & FastaStreamIterator::operator++()
{
    pimpl_->advance();
    return *this;
}

// --------------------------------------------------------------------------

FastaStreamIterator FastaStreamIterator::operator++(int)
{
    FastaStreamIterator t(*this);
    ++(*this);
    return *this;
}

// --------------------------------------------------------------------------

bool FastaStreamIterator::operator==(FastaStreamIterator const & rhs) const
{
    return pimpl_->equal(rhs.pimpl_.get());
}

// --------------------------------------------------------------------------

bool FastaStreamIterator::operator!=(FastaStreamIterator const & rhs) const
{
    return !(*this == rhs);
}


// --------------------------------------------------------------------------

} // namespace swift
