// gram.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#include "gram.hpp"

namespace swift {

// --------------------------------------------------------------------------

/*!
    Constructs an empty gram of size \a size with mask sensitivty \a ms.
    If \a ms is \l swift::MaskSensitive, the gram will become invalid
    over masked letters.
*/
Gram::Gram(unsigned int size, MaskSensitivity ms) :
    size_(size),
    value_(0),
    flags_(0),
    valueMask_(valueMask(size)),
    flagMask_(flagMask(size, ms)),
    frontShift_((size - 1) << 1)
{
    clear();
}


/*!
    Destroys the Gram object.
*/
Gram::~Gram()
{
}


/*!
    Returns the value mask for grams of size \a size.
*/
Gram::Value Gram::valueMask(unsigned int size)
{
    return (1UL << (size << 1)) - 1;
}


/*!
    Returns the flag mask for grams of size \a size with mask sensitivity
    \a ms.
*/
Gram::Value Gram::flagMask(unsigned int size, MaskSensitivity ms)
{
    return ms == MaskSensitive ?
           valueMask(size) : (valueMask(size) & 0xaaaaaaaa);
}


/*!
    Clears the gram. The resulting empty gram is invalid.

    \sa isValid()
*/
void Gram::clear()
{
    value_ = 0;
    flags_ = flagMask_;
}


/*!
    Returns the maximum numerical value for a grams of size \a size.
*/
Gram::Value Gram::maxValue(unsigned int size)
{
    return valueMask(size);
}

// --------------------------------------------------------------------------

} // namespace swift

