// gramchainaligner.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#ifndef _SWIFT_GRAMCHAINALIGNER_
#define _SWIFT_GRAMCHAINALIGNER_

#include "gram.hpp"
#include "alignment.hpp"
#include "assert.hpp"
#include "hashmap.hpp"
#include "masksensitivity.hpp"
#include "scorefunction.hpp"
#include <iostream>
#include <map>
#include <stdexcept>
#include <vector>


namespace swift {


namespace gca_ {

typedef struct {
    unsigned int gramSize;
    char const *    query;          ///< Query sequence.
    char const *    subject;        ///< Subject sequence.
    int            queryLength;    ///< Length of query sequence.
    int            subjectLength;  ///< Length of subject sequence.
    int            lowerDiagonal;  ///< Lower diagonal of band.
    int            upperDiagonal;  ///< Upper diagonal of band.
    unsigned int           minHits;        ///< Minimum number of hits in chain.
    int            maxGap;         ///< Max gap between successive hits in chain.
} Parameters;

struct Position {
    Position * next;
    unsigned int      position;
};

class PositionPool
{
    public:
    PositionPool(void);
    ~PositionPool(void);

    /// Reserve space for poolSize positions
    /// and clear the pool.
    void reserve(unsigned int poolSize);

    /// Push position back to the pool.
    void push(Position * position);

    /// Pop position off pool.
    Position * pop(void);

    /// 
    void clear(void);
    
    private:
    Position * pool_;
    Position * freeEntries_;
    unsigned int      poolSize_;
    unsigned int      unusedEntries_;
};

struct PositionList {
    Position * first;
    Position * last;

    inline PositionList() :
        first(0), last(0) {};
    inline PositionList(Position * f, Position * l) :
        first(f), last(l) {};
};

struct Link {
    unsigned int x;                ///< X coordinate of link (begin).
    int  y;                ///< Y coordinate of link (begin).
    int  score;            ///< Score of chain ending in this link.
    unsigned int length;           ///< Length of ungapped segment.
    union {
        unsigned int hits;         ///< Number of hits in chain ending in this link.
        unsigned int clusterFlag;  ///< Zero if link has been clustered, otherwise non-zero.
    };
    union {
        unsigned int refCount;     ///< Number of links linking to this link.
        unsigned int clusterNum;   ///< Cluster number.
    };
    Link * next;            ///< Next (previous) link in chain.
};

class LinkPool
{
    public:
    static unsigned int const DefaultBlockSize;
    
    /// Constructs a link pool.
    LinkPool(void);

    /// Constructs a link pool using a specific block size.
    LinkPool(unsigned int blockSize);

    /// Destructor.
    virtual ~LinkPool(void);
   
    /// 
    void clear(void);

    /// Dispose a chain of links.
    void disposeLinks(Link * link);

    /// Acquire a link.
    Link * acquireLink(void);

    // Return number of links in 
    unsigned int getBlockSize() const;
    unsigned int getBlockCount() const;

    private:
    /// Initialize
    void initialize(void);

    struct LinkBlock {
        LinkBlock * next;
        Link *      links;
    };

    unsigned int blockSize_;
    unsigned int blockCount_;
    
    LinkBlock * usedBlocks_;
    LinkBlock * unusedBlocks_;

    Link * unusedLinks_;
    unsigned int  unusedLinksCount_;
    Link * recycledLinks_;
};

} // namespace gca_


namespace {
    using namespace gca_;
}



class GramChainAligner
{
    private:
    struct GramHash
    {
        inline unsigned int operator()(Gram::Value key) const {
            return (key * 2654435761UL); // D. Knuth, golden ratio.
        }
    };

    struct GramCompare
    {
        inline bool operator()(
            Gram::Value const & A,
            Gram::Value const & B
        ) const
        {
            return (A == B);
        }
    };

    typedef HashMap<
                Gram::Value,
                PositionList,
                GramCompare,
                GramHash
            > GramPositionMap;

    typedef GramPositionMap::value_type GramPosition;

    typedef std::map<int, Link *>     LinkList;
    typedef std::vector<LinkList *>    LinkListPool;
    typedef std::map<int, LinkList *> DiagonalMap;
    typedef std::vector<Link *>        ChainVector;


    public:
    typedef gca_::Parameters Parameters;
    typedef std::vector<Alignment *> Result;

    public:
    GramChainAligner();
    ~GramChainAligner();
    
    void alignments(
        Parameters const & params,
        Result & result
    );

    /// Finds alignments with edit scripts.
    void alignWithEditScript(
        Parameters const & params,
        Result & result
    );

    void setScoreFunction(ScoreFunction const & scoreFunction);
    void setMatchScore(int matchScore);

    void setQueryMaskSensitivity(MaskSensitivity ms);
    MaskSensitivity queryMaskSensitivity() const;

    void setSubjectMaskSensitivity(MaskSensitivity ms);
    MaskSensitivity subjectMaskSensitivity() const;

    private:
    /// Find chains of q-grams in a parallelogram.
    void findChains(
        char const *  query,
        char const *  subject,
        int          queryLength,
        int          subjectLength,
        int          lowerDiagonal,
        int          upperDiagonal,
        int          maxGap,
        unsigned int         minChainLength,
        unsigned int gramSize
    ); 
 
    /// Rescore chains into alignments.
    void alignmentsFromChains(
        Parameters const & params,
        Result & result
    );

    void rescoreOnDiagonal(
        Link const *       plink,
        Link const *       link,
        Parameters const & params,
        int &             score,
        EditScript *       editScript
    );

    void rescoreOffDiagonal(
        Link const *       plink,
        Link const *       link,
        Parameters const & params,
        int &             score,
        EditScript *       editScript
    );

    void showChain(
        char const * query,
        char const * subject,
        Link *       chain,
        unsigned int gramSize
    );

    private:
    GramPositionMap gramMap_;       ///< Hash map of gram positions.
    
    PositionPool    positionPool_;  ///< Pool of gram positions.
    LinkPool        linkPool_;      ///< Pool of links.
    LinkListPool    linkListPool_;  ///< Pool of link lists.
    DiagonalMap     diagonalMap_;   ///< Map of diagonals to link lists.

    ChainVector     chainsFound_;   ///< Vector of found chains.

    MaskSensitivity queryMaskSensitivity_;
    MaskSensitivity subjectMaskSensitivity_;

    ScoreFunction scoreFunction_;
    int matchScore_;
};



namespace gca_ {

/* ------------------------------------------------------------------------ */

inline
void PositionPool::push(Position * position)
{
    position->next = freeEntries_;
    freeEntries_   = position;
}

/* ------------------------------------------------------------------------ */

inline
Position * PositionPool::pop(void)
{
    if (freeEntries_) {
        Position * t(freeEntries_);
        freeEntries_ = freeEntries_->next;
        t->next = 0;
        return t;
    }

    if (unusedEntries_) {
        Position * t(pool_ + --unusedEntries_);
        t->next = 0;
        return t;
    }

    throw std::length_error("PositionPool: Out of entries!");
}
/* ------------------------------------------------------------------------ */
    
inline
Link * LinkPool::acquireLink(void)
{
	// Attempt acquiring a recycled link
	if (recycledLinks_) {
		Link * t(recycledLinks_);
		recycledLinks_ = t->next;
        return (t);
	}

	// Attempt acquiring an unused link
	if (unusedLinksCount_) {
		--unusedLinksCount_;
		return (unusedLinks_++);
	}
	
	// Check for an unused block
	if (unusedBlocks_) {
		LinkBlock * b(unusedBlocks_);
		unusedBlocks_ = unusedBlocks_->next;
		b->next       = usedBlocks_;
		usedBlocks_   = b;

		unusedLinks_      = b->links + 1;
		unusedLinksCount_ = blockSize_ - 1;
		return (b->links);
	}
	
	// No unused blocks, make a new
	LinkBlock * newBlock(new LinkBlock());
	newBlock->links   = new Link[blockSize_];
	newBlock->next    = usedBlocks_;
	usedBlocks_       = newBlock;
	unusedLinks_      = newBlock->links + 1;
	unusedLinksCount_ = blockSize_ - 1;
    ++blockCount_;

	return (newBlock->links);
}

/* ------------------------------------------------------------------------ */

inline
void LinkPool::disposeLinks(Link * link)
{
#ifndef NDEBUG
    // FIXME 1148916879
    std::cerr << "LinkPool::disposeLinks() not implemented!!!\n";
#endif

    return;

#if 0
    Link * first(link);
    Link * last(0);

    while (link && !link->refCount) {
        last = link;
        if (!link->next)
            break;
        --(link->next->refCount);
        link = link->next;
    }

    last->next = recycledLinks_;
    recycledLinks_ = first;
#endif
}

/* ------------------------------------------------------------------------ */

inline
void LinkPool::clear(void)
{
    unusedLinksCount_ = 0;
    recycledLinks_    = 0;

    if (!unusedBlocks_) {
        unusedBlocks_ = usedBlocks_;
    } else {
        LinkBlock * t(unusedBlocks_);
        while (t->next)
            t = t->next;
        t->next = usedBlocks_;
    }
    usedBlocks_ = 0;
}


} // namespace gca_


} // namespace swift

#endif


