// gramindex.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#ifndef _SWIFT_GRAMINDEX_
#define _SWIFT_GRAMINDEX_

#include "gram.hpp"
#include "masksensitivity.hpp"
#include "sequencedatabase.hpp"
#include <boost/range.hpp>
#include <boost/shared_array.hpp>
#include <boost/utility.hpp>

namespace swift {

// --------------------------------------------------------------------------

class GramIndex : public boost::noncopyable
{
    public:
    typedef SequenceDatabase::size_type size_type; 
    typedef SequenceDatabase::size_type Offset; 
    typedef SequenceDatabase::size_type Position;
    typedef boost::iterator_range<Position const *> Positions;

    static GramIndex * construct(
        SequenceDatabase const * database,
        unsigned int gramSize,
        MaskSensitivity maskSensitivity
    );
    
    Positions positions(Gram::Value gramValue) const;

    bool serialize(std::string const & name) const;
    static GramIndex * unserialize(
        std::string const & name,
        unsigned int gramSize
    );

    unsigned int gramSize() const;
    size_type gramCount() const;
    Position maxPosition() const;
    MaskSensitivity maskSensitivity() const;

    private:
    struct Header;

    GramIndex(
        Offset const * lookup,
        Position const * positions,
        Position maxPosition,
        unsigned int gramSize,
        MaskSensitivity maskSensitivity
    );

    GramIndex(
        boost::shared_array<Offset const> lookup,
        boost::shared_array<Position const> positions,
        Position maxPosition,
        unsigned int gramSize,
        MaskSensitivity maskSensitivity
    );
        
    void serializeHeader(std::ostream & outputStream) const;
    void serializeLookup(std::ostream & outputStream) const;
    void serializePositions(std::ostream & outputStream) const;

    static std::string lookupFile(
        std::string const & name,
        unsigned int gramSize
    );

    static std::string positionsFile(
        std::string const & name,
        unsigned int gramSize
    );

    unsigned int gramSize_;
    MaskSensitivity maskSensitivity_;

    boost::shared_array<Offset const> lookup_;
    boost::shared_array<Position const> positions_;
    Position maxPosition_;

    static char const * LookupSuffix;
    static char const * PositionsSuffix;
};

// --------------------------------------------------------------------------

/*!
    Returns the size of the indexed grams.
*/
inline
unsigned int GramIndex::gramSize() const
{
    return gramSize_;
}


/*!
    Returns the number of indexed grams.
*/
inline
GramIndex::size_type GramIndex::gramCount() const
{
    return lookup_[Gram::maxValue(gramSize_) + 1];
}


/*!
    Returns the position of the last indexed gram.
*/
inline
GramIndex::Position GramIndex::maxPosition() const
{
    return maxPosition_;
}


/*!
    Returns the mask sensitivity used for construction of the gram index.
*/
inline
MaskSensitivity GramIndex::maskSensitivity() const
{
    return maskSensitivity_;
}


/*!
    Returns an iterator range for the positions in the sequence database
    where a gram with value \a gramValue begins.
*/
inline
GramIndex::Positions GramIndex::positions(Gram::Value gramValue) const
{
    return boost::make_iterator_range(
        positions_.get() + lookup_[gramValue],
        positions_.get() + lookup_[gramValue + 1]
    );
}

// --------------------------------------------------------------------------

} // namespace swift

#endif
