// gramview.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#ifndef _SWIFT_GRAMVIEW_
#define _SWIFT_GRAMVIEW_

#include "gram.hpp"
#include <boost/cstdint.hpp>
#include <iterator>

namespace swift {

// --------------------------------------------------------------------------

template <typename DI>
class GramView
{
    public:
    typedef DI                 domain_iterator;
    typedef Gram               value_type;
    typedef value_type *       pointer;
    typedef const value_type * const_pointer;
    typedef value_type &       reference;
    typedef const value_type & const_reference;
    typedef std::size_t        size_type;
    
    class const_iterator :
        public std::iterator<std::bidirectional_iterator_tag, value_type> 
    {
        private:
        const_iterator();
        const_iterator(
            domain_iterator position,
            domain_iterator end,
            Gram const & gram
        );

        public:
        const_reference operator*() const;
        const_pointer operator->() const;

        const_iterator & operator++();
        const_iterator operator++(int);
        const_iterator & operator--();
        const_iterator operator--(int);

        bool operator==(const_iterator const & other) const;
        bool operator!=(const_iterator const & other) const;

        private:
        domain_iterator position_;
        domain_iterator end_;
        Gram gram_;
        
        friend class GramView;
    };

    GramView(
        domain_iterator begin,
        domain_iterator end,
        unsigned int gramSize,
        MaskSensitivity ms
    );
    ~GramView() {}

    const_iterator begin() const;
    const_iterator end() const;
    size_type size() const;
    
    unsigned int gramSize() const;
    MaskSensitivity maskSensitivity() const;

    private:
    unsigned int gramSize_;
    size_type size_;
    domain_iterator begin_;
    domain_iterator end_;
    MaskSensitivity maskSensitivity_;
};

// --------------------------------------------------------------------------

/*!
    Constructs a GramView for grams of size \a gramSize over the range
    [\a begin, \a end). If \a ms is \l swift::MaskSensitive, grams will
    be invalid when iterating over a masked part of the range.
*/
template <typename DI>
inline
GramView<DI>::GramView(
    domain_iterator begin,
    domain_iterator end,
    unsigned int gramSize,
    MaskSensitivity ms
) :
    gramSize_(gramSize),
    begin_(begin),
    end_(end),
    maskSensitivity_(ms)
{
    if (begin_ >= end_) {
        size_ = 0;
    } else {
        size_ = end_ - begin_ < gramSize_ ? 0 : end_ - begin_ - gramSize_ + 1;
    }
}


/*!
    Returns the number of grams in the gram view. This number does not
    take validity of the grams into account.
*/
template <typename DI>
inline
typename GramView<DI>::size_type GramView<DI>::size() const
{
    return size_;
}


/*!
    Returns the size of the grams in the gram view.
*/
template <typename DI>
inline
unsigned int GramView<DI>::gramSize() const
{
    return gramSize_;
}


/*!
    Returns the mask sensitivity of the gram view.

    \sa swift::MaskSensitivity
*/
template <typename DI>
inline
MaskSensitivity GramView<DI>::maskSensitivity() const
{
    return maskSensitivity_;
}


/*!
    Returns a gram iterator to the beginning of the view range.
*/
template <typename DI>
inline
typename GramView<DI>::const_iterator GramView<DI>::begin() const
{
    Gram gram(gramSize_, maskSensitivity_);
    if (size()) {
        gram.push_back(begin_, begin_ + gramSize_);
        return const_iterator(begin_ + gramSize_ - 1, end_, gram);
    } else {
        return const_iterator(end_, end_, gram);
    }
}


/*!
    Returns a gram iterator to the end of the view range.
*/
template <typename DI>
inline
typename GramView<DI>::const_iterator GramView<DI>::end() const
{
    Gram gram(gramSize_, maskSensitivity_);
    if (size()) {
        gram.push_back(end_ - gramSize_ + 1, end_ - 1);
        gram.push_back();
    }
    return const_iterator(end_, end_, gram);
}

// --------------------------------------------------------------------------

/*!
    Constructs a gram iterator pointing to position \a position, with gram
    \a gram as current value. The \a end argument specifies the past-end
    position in the parenting GramView.

    \internal
    The past-end position is needed due to the special semantic of the
    increment operator. 
*/
template <typename DI>
inline
GramView<DI>::const_iterator::const_iterator(
    domain_iterator position,
    domain_iterator end,
    Gram const & gram
) :
    position_(position),
    end_(end),
    gram_(gram)
{
}


/*!
    Returns a const reference to the current gram.
*/
template <typename DI>
inline
typename GramView<DI>::const_reference
GramView<DI>::const_iterator::operator*() const
{
    return gram_;
}


/*!
    Returns a const pointer to the current gram.
*/
template <typename DI>
inline
typename GramView<DI>::const_pointer
GramView<DI>::const_iterator::operator->() const
{
    return &gram_;
}


/*!
    Preincrements the gram iterator.

    \internal
    To the client, the semantic is STL-like. However, the increment operation
    in fact both increments and dereferences. It does both to make the gram
    value available after multiple increments and no dereferences.
    To avoid dereferencing end() of the parent GramView we must check the
    validity of the position.

    Note that the checking can be alleviated by always setting the last letter
    of the gram in operator*() and shifting the gram in operator++(). Checking
    the end position is however sligthly more efficient.
*/
template <typename DI>
inline
typename GramView<DI>::const_iterator &
GramView<DI>::const_iterator::operator++()
{
    if (++position_ < end_) {
        gram_.push_back(*position_);
    }
    return *this;
}


/*!
    Postincrements the gram iterator.
*/
template <typename DI>
inline
typename GramView<DI>::const_iterator
GramView<DI>::const_iterator::operator++(int)
{
    const_iterator returnValue(*this);
    ++(*this);
    return returnValue;
}


/*!
    Predecrements the gram iterator.
*/
template <typename DI>
inline
typename GramView<DI>::const_iterator &
GramView<DI>::const_iterator::operator--()
{
    gram_.push_front(*(--position_ - gram_.size() + 1));
    return *this;
}


/*!
    Postdecrements the gram iterator.
*/
template <typename DI>
inline
typename GramView<DI>::const_iterator
GramView<DI>::const_iterator::operator--(int)
{
    const_iterator returnValue(*this);
    --(*this);
    return returnValue;
}


/*!
    Compares for equivalence with other gram iterator. Returns true if both
    iterators point to the same position.
*/
template <typename DI>
inline
bool
GramView<DI>::const_iterator::operator==(const_iterator const & other) const
{
    return (position_ == other.position_);
}


/*!
    Compares for inequivalence with other gram iterator. Returns true if the
    iterators point to different positions.
*/
template <typename DI>
inline
bool
GramView<DI>::const_iterator::operator!=(const_iterator const & other) const
{
    return !(*this == other);
}

// --------------------------------------------------------------------------

} // namespace swift

#endif

