// idmap.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#include "idmap.hpp"
#include "binary_ostream_iterator.hpp"
#include <cstring>
#include <algorithm>

namespace swift {

// IdMapBase class definition -----------------------------------------------

uint32 IdMapBase::hash(std::string const & key) const
{
    uint32 h(5381);
    std::string::const_iterator i(key.begin()), j(key.end());
    for ( ; i < j; ++i) {
        h *= 33;
        h ^= *i;
    }
    return h;
}

// IdMap class definition ---------------------------------------------------

std::pair<bool, uint32> IdMap::find(std::string const & key) const
{
    uint32 hashKey(hash(key));
    uint32 mapGroup(hashKey % MapGroupCount);
    
    if (mapGroupOffsets_[mapGroup] == mapGroupOffsets_[mapGroup + 1])
        return std::pair<bool, uint32>(false, 0);

    MapEntry v(hashKey, 0, 0, 0);
    std::vector<MapEntry>::const_iterator gi = 
        mapEntries_.begin() + mapGroupOffsets_[mapGroup];
    std::vector<MapEntry>::const_iterator gj =
        mapEntries_.begin() + mapGroupOffsets_[mapGroup + 1];

    std::pair<
        std::vector<MapEntry>::const_iterator,
        std::vector<MapEntry>::const_iterator
    > r = std::equal_range(gi, gj, v);

    for ( ; r.first != r.second; ++r.first) {
        MapEntry const & e = *(r.first);
        if (strncmp(key.c_str(), mapKeys_ + e.keyOffset_, e.keySize_) == 0)
            return std::pair<bool, uint32>(true, e.id_);
    }
    return std::pair<bool, uint32>(false, 0);
}


uint32 IdMap::size() const
{
    return mapEntries_.size();
}


IdMap::IdMap(
    char * mapKeys,
    std::vector<uint32> & mapGroupOffsets,
    std::vector<MapEntry> & mapEntries
) :
    mapKeys_(mapKeys)
{
    mapGroupOffsets_.swap(mapGroupOffsets);
    mapEntries_.swap(mapEntries);

#if 0
    std::string key;
    for (size_t i(0), j(mapEntries_.size()); i < j; ++i) {
        key = std::string(mapKeys_ + mapEntries_[i].keyOffset_,
                          mapEntries_[i].keySize_);
        std::cout << "#" << i << " (" << mapEntries_[i].hashKey_ 
                  << " / " << mapEntries_[i].id_ << ") = "
                  << key << "\n";
    }
    for (size_t i(0), j(mapGroupOffsets_.size()); i < j; ++i) {
        std::cout << i << ": " << mapGroupOffsets_[i] << "\n";
    }
#endif
}


IdMap::~IdMap()
{
    delete [] mapKeys_;
}


// IdMapFactory::Source class definition ------------------------------------

IdMap * IdMapFactory::Source::read(std::istream & input)
{
    readKeys(input);
    readMapGroupOffsets(input);
    readMapEntries(input);
    IdMap * idMap = new IdMap(mapKeys_, mapGroupOffsets_, mapEntries_);
    cleanUp();
    return idMap;
}


void IdMapFactory::Source::cleanUp()
{
    mapKeys_ = 0;
    std::vector<uint32>().swap(mapGroupOffsets_);
    std::vector<MapEntry>().swap(mapEntries_);
}


void IdMapFactory::Source::readKeys(std::istream & is)
{
    uint32 totalKeyLength(0);
    is.read((char *)&totalKeyLength, sizeof(totalKeyLength));
    mapKeys_ = new char[totalKeyLength];
    is.read((char *)mapKeys_, totalKeyLength);
}


void IdMapFactory::Source::readMapGroupOffsets(std::istream & is)
{
    uint32 offset(0);
    mapGroupOffsets_.reserve(MapGroupCount + 1);
    for (size_t i(0), j(MapGroupCount + 1); i < j; ++i) {
        is.read((char *)&offset, sizeof(offset));
        mapGroupOffsets_.push_back(offset);
    }
}


void IdMapFactory::Source::readMapEntries(std::istream & is)
{
    MapEntry e;
    mapEntries_.reserve(mapGroupOffsets_.back());
    for (size_t i(0),j(mapGroupOffsets_.back()); i < j; ++i) {
        is.read((char *)&e, sizeof(e));
        mapEntries_.push_back(e);
    }
}

// IdMapFactory::Sink class definition --------------------------------------

IdMapFactory::Sink::Sink() :
    mapEntryCount_(0)
{
initialize();
}


IdMapFactory::Sink::~Sink()
{
    cleanUp();
}


void IdMapFactory::Sink::clear()
{
    cleanUp();
}


void IdMapFactory::Sink::initialize()
{
    mapEntries_.reserve(MapGroupCount);
    for (size_t i(0); i < MapGroupCount; ++i)
        mapEntries_.push_back(new std::vector<MapEntry>);
}


void IdMapFactory::Sink::cleanUp()
{
    if (mapEntryCount_) {
        for (size_t i(0); i < MapGroupCount; ++i)
            delete mapEntries_[i];
        std::vector< std::vector<MapEntry> *>().swap(mapEntries_);
        std::string().swap(mapKeys_);
        mapEntryCount_ = 0;
    }
}


void IdMapFactory::Sink::write(
    std::string const & key,
    uint32 id 
)
{
    if (!mapEntryCount_)
        initialize();

    MapEntry e(hash(key), mapKeys_.size(), key.size(), id);
    mapEntries_[e.hashKey_ % MapGroupCount]->push_back(e);
    mapKeys_.append(key);
    ++mapEntryCount_;
}


void IdMapFactory::Sink::write(std::ostream & output)
{
    sortMapGroups();
    
    writeKeys(output);
    writeMapGroupOffsets(output);
    writeMapEntries(output);
}


void IdMapFactory::Sink::writeKeys(std::ostream & os)
{
    uint32 totalKeyLength(mapKeys_.size());
    os.write((char *)&totalKeyLength, sizeof(totalKeyLength));
    os.write(mapKeys_.data(), mapKeys_.size());
}


void IdMapFactory::Sink::writeMapGroupOffsets(std::ostream & os)
{
    uint32 offset(0);
    for (size_t i(0), j(MapGroupCount); i < j; ++i) {
        os.write((char *)&offset, sizeof(offset));
        offset += mapEntries_[i]->size();
    }
    os.write((char *)&offset, sizeof(offset)); // sentinel
}


void IdMapFactory::Sink::writeMapEntries(std::ostream & os)
{
    for (size_t i(0), j(MapGroupCount); i < j; ++i) {
        std::copy(
            mapEntries_[i]->begin(),
            mapEntries_[i]->end(),
            binary_ostream_iterator<MapEntry>(os)
        );
    }
}


void IdMapFactory::Sink::sortMapGroups()
{
    for (size_t i(0), j(MapGroupCount); i < j; ++i) {
        std::sort(mapEntries_[i]->begin(), mapEntries_[i]->end());
    }
}


} // namespace swift


