// idmap.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//
// Additions / changes by Peter Husemann
//

#ifndef _SWIFT_IDMAP_
#define _SWIFT_IDMAP_

#include <boost/cstdint.hpp>
#include <iosfwd>
#include <string>
#include <vector>

namespace swift {

typedef boost::uint32_t uint32;

// IdMapBase class declaration ----------------------------------------

class IdMapBase
{
    protected:
    enum { MapGroupCount = 256 };

    struct MapEntry
    {
        MapEntry() : hashKey_(0), keyOffset_(0), keySize_(0), id_(0) { }
        MapEntry(uint32 hashKey, uint32 keyOffset, uint32 keySize, uint32 id) :
            hashKey_(hashKey), keyOffset_(keyOffset),
            keySize_(keySize), id_(id) { }

        bool operator<(MapEntry const & other) const {
            return hashKey_ < other.hashKey_;
        }

        uint32 hashKey_;
        uint32 keyOffset_;
        uint32 keySize_;
        uint32 id_;
    };

    virtual ~IdMapBase() { }
    uint32 hash(std::string const & key) const;
};

// IdMapFactory class declaration -------------------------------------------

class IdMap;

class IdMapFactory
{
    public:
    class Source : private IdMapBase
    {
        public:
        Source() { }
        IdMap * read(std::istream & input);

        private:
        void readKeys(std::istream & is);
        void readMapGroupOffsets(std::istream & is);
        void readMapEntries(std::istream & is);
        void cleanUp();

        char * mapKeys_;
        std::vector<uint32> mapGroupOffsets_;
        std::vector<MapEntry> mapEntries_;
    };

    class Sink : private IdMapBase
    {
        public:
        Sink();
        ~Sink();
        void clear();
        void write(std::string const & key, uint32 id);
        void write(std::ostream & output);

        private:
        void initialize();
        void cleanUp();
        void sortMapGroups();
        void writeKeys(std::ostream & os);
        void writeMapGroupOffsets(std::ostream & os);
        void writeMapEntries(std::ostream & os);

        std::string mapKeys_;
        std::vector< std::vector<MapEntry> *> mapEntries_;
        uint32 mapEntryCount_;
    };
};

// IdMap class declaration --------------------------------------------------

class IdMap : private IdMapBase
{
    public:
    std::pair<bool, uint32> find(std::string const & key) const;
    uint32 size() const;

    public:
    ~IdMap();

    private:
    IdMap(
        char * mapKeys,
        std::vector<uint32> & mapGroupOffsets,
        std::vector<MapEntry> & mapEntries
    );

    char * mapKeys_;
    std::vector<uint32> mapGroupOffsets_;
    std::vector<MapEntry> mapEntries_;

    struct Compare {
        bool operator()(MapEntry const & entry, uint32 hashKey) const {
            return entry.hashKey_ == hashKey;
        }
    };

    friend class IdMapFactory::Source;
};

// --------------------------------------------------------------------------

} // namespace swift

#endif




