// karlinaltschul.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#ifndef _SWIFT_KARLINALTSCHUL_
#define _SWIFT_KARLINALTSCHUL_

#include <cstdlib>
#include <cmath>

namespace swift
{


namespace {
    using std::abs;
    using std::log;
    using std::exp;
};


class KarlinAltschul
{
    public:
    KarlinAltschul(int matchScore, int mismatchScore);
    
    double eValue(
        int score,
        unsigned int querySize,
        unsigned int databaseSize
    ) const;

    double bitScore(int score) const;
    int rawScore(double bitScore) const;

    private:
    struct KarlinAltschulParameters
    {
        double lambda;
        double K;
        double H;
    };

    KarlinAltschulParameters parameters_;

    static const KarlinAltschulParameters KaParameters[10][10];
};


/*!
    Construct a new KarlinAltschul object for the match and mismatch scores,
    \a matchScore and \a mismatchScore, respectively.
*/
inline
KarlinAltschul::KarlinAltschul(int matchScore, int mismatchScore) :
    parameters_(KaParameters[abs(mismatchScore) - 1][abs(matchScore) - 1])
{
}


/*!
    Returns the e-value for a given \a score, query size and database size.
*/
inline
double KarlinAltschul::eValue(
    int score,
    unsigned int querySize,
    unsigned int databaseSize
) const
{
    return (parameters_.K * querySize * databaseSize *
            exp(-parameters_.lambda * score));
}


/*!
    Returns the bit score for a given \a score.
*/
inline
double KarlinAltschul::bitScore(int score) const
{
    static const double Log2(log(2.0));
    
    return (parameters_.lambda * (double(score)) - log(parameters_.K))/Log2;
}


/*!
    Returns the raw score corresponding to a given bit score.
*/
inline
int KarlinAltschul::rawScore(double bitScore) const
{
    static const double Log2(log(2.0));
    
    return static_cast<int>(
        (bitScore*Log2 + log(parameters_.K)) / parameters_.lambda
    );
}


} // namespace swift

#endif

