// scorematrix.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//  

#include "scorematrix.hpp"
#include "encoding.hpp"
#include <algorithm>

namespace swift
{


unsigned int ScoreMatrix::MatrixSize = 
    Encoding::ValueCount * Encoding::ValueCount;


/*!
    Constructs a score matrix where all entries are zero.
*/
ScoreMatrix::ScoreMatrix()
{
    matrix_ = new int[MatrixSize];
    std::fill(matrix_, matrix_ + MatrixSize, 0);
}


/*!
    Constructs a score matrix where matches are assigned the score \a
    matchScore and mismatches are assigned the score \a mismatchScore. If \a ms
    is MaskSensitive, the upper and lower case of a symbol generate a mismatch.
*/
ScoreMatrix::ScoreMatrix(
    int matchScore,
    int mismatchScore,
    MaskSensitivity ms
)
{
    matrix_ = new int[MatrixSize];
    std::fill(matrix_, matrix_ + MatrixSize, mismatchScore);

    for (unsigned int i(0), j(4); i < j; ++i) {
        setEntry(i, i, matchScore);
        setEntry(i + 4, i + 4, matchScore);
    }

    if (ms == MaskInsensitive) {
        for (unsigned int i(0), j(4); i < j; ++i) {
            setEntry(i, i + 4, matchScore);
            setEntry(i + 4, i, matchScore);
        }
    }
}


/*!
    Constructs a copy of the score matrix \a other.
*/
ScoreMatrix::ScoreMatrix(ScoreMatrix const & other)
{
    matrix_ = new int[MatrixSize];
    std::copy(other.matrix_, other.matrix_ + MatrixSize, matrix_);
}


/*!
    Destroys the ScoreMatrix object.
*/
ScoreMatrix::~ScoreMatrix()
{
    delete [] matrix_;
}


/*!
    Assignment operator.
*/
ScoreMatrix & ScoreMatrix::operator=(ScoreMatrix const & other)
{
    if (this == &other)
        return *this;

    std::copy(other.matrix_, other.matrix_ + MatrixSize, matrix_);
    return *this;
}


} // namespace swift

