// sequence.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#ifndef _SWIFT_SEQUENCE_
#define _SWIFT_SEQUENCE_

#include <boost/shared_array.hpp>
#include <stdexcept>
#include <iostream>

namespace swift { 

// --------------------------------------------------------------------------

class Sequence
{
    public:
    static size_t const npos;

    typedef unsigned char       value_type;
    typedef value_type *        pointer;
    typedef value_type const *  const_pointer;
    typedef pointer             iterator;
    typedef const_pointer       const_iterator;
    typedef value_type &        reference;
    typedef const value_type &  const_reference;
    typedef size_t              size_type;
    typedef ptrdiff_t           difference_type;

    typedef boost::shared_array<value_type> Buffer;

    Sequence();
    Sequence(Sequence const & other, size_type pos = 0, size_type n = npos);
    Sequence(Buffer const & buffer, size_type size, size_type offset = 0);
    Sequence(char const * string, size_type size);
    
    template <class InputIterator>
    Sequence(InputIterator first, InputIterator last);

    iterator begin();
    iterator end();
    const_iterator begin() const;
    const_iterator end() const;

    reference operator[](unsigned int i);
    const_reference operator[](unsigned int i) const;

    pointer data();
    const_pointer const_data() const;

    size_type size() const;
    bool empty() const;

    void reverseComplement();

    private:
    void accessCow(bool markUnshareable = false);
    void cloneBuffer();
    void cloneBuffer(const_pointer first, const_pointer last);

    Buffer  buffer_;
    pointer begin_;
    pointer end_;
    bool shareable_; // TODO 1147704578
};

// --------------------------------------------------------------------------

template <class InputIterator>
Sequence::Sequence(InputIterator first, InputIterator last) :
    buffer_(0),
    begin_(0),
    end_(0),
    shareable_(true)
{
    cloneBuffer(first, last);
}


inline
Sequence::iterator Sequence::begin()
{
    accessCow(true);
    return begin_;
}


inline
Sequence::iterator Sequence::end()
{
    accessCow(true);
    return end_;
}


inline
Sequence::const_iterator Sequence::begin() const
{
    return begin_;
}


inline
Sequence::const_iterator Sequence::end() const
{
    return end_;
}


inline
Sequence::size_type Sequence::size() const
{
    return end_ - begin_;
}


inline
bool Sequence::empty() const
{
    return !size();
}


inline
Sequence::reference Sequence::operator[](unsigned int i)
{
    accessCow(true);
    return *(begin_ + i);
}


inline
Sequence::const_reference Sequence::operator[](unsigned int i) const
{
    return *(begin_ + i);
}


inline
Sequence::pointer Sequence::data()
{
    accessCow(true);
    return begin_;
}


inline
Sequence::const_pointer Sequence::const_data() const
{
    return begin_;
}


inline
void Sequence::accessCow(bool markUnshareable)
{
    if (!buffer_.unique())
        cloneBuffer();

    if (markUnshareable)
        shareable_ = false;
}


inline
void Sequence::cloneBuffer()
{
    cloneBuffer(begin_, end_);
}


inline
void Sequence::cloneBuffer(const_pointer first, const_pointer last)
{
    size_type sz(last - first);
    buffer_.reset(new value_type[sz]);
    begin_ = buffer_.get();
    end_   = begin_ + sz;
    std::copy(first, last, begin_);
}


} // namespace swift

#endif

