// streamutil.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#include "streamutil.hpp"
#include <cassert>
#include <cstdlib>
#include <cstring>

namespace swift {

namespace detail {

}

// Function definition ------------------------------------------------------

void pushback(
    std::istream & is,
    std::char_traits<char>::char_type * buf,
    std::streamsize buf_size,
    void * del_ptr
)
{
    PushbackStreamBuf * sb = dynamic_cast<PushbackStreamBuf *> (is.rdbuf());

    if (sb) {
        if (sb->m_Buf <= buf && buf + buf_size == sb->gptr()) {
            sb->setg(buf, buf, sb->m_Buf + sb->m_BufSize);
            return;
        }

        if ((std::streamsize)(sb->gptr() - sb->m_Buf) >= buf_size) {
            std::char_traits<char>::char_type* bp = sb->gptr() - buf_size;
            if (::memcmp(buf, bp, buf_size) == 0) {
                sb->setg(bp, bp, sb->egptr());
                if ( del_ptr ) {
                    delete[] (std::char_traits<char>::char_type*) del_ptr;
                }
                return;
            }
        }
    }

    (void) new PushbackStreamBuf(is, buf, buf_size, del_ptr);
}

// --------------------------------------------------------------------------

void pushback(
    std::istream & is,
    const std::char_traits<char>::char_type * buf,
    std::streamsize buf_size
)
{
    std::char_traits<char>::char_type * buf_copy = new std::char_traits<char>::char_type[buf_size];
    ::memcpy(buf_copy, buf, buf_size);
    pushback(is, buf_copy, buf_size, buf_copy);
}

// --------------------------------------------------------------------------

std::streamsize readsome(
    std::istream & is,
    std::char_traits<char>::char_type * buf,
    std::streamsize buf_size
)
{
    // FIXME 1089121662 
    
    std::streamsize avail(is.rdbuf()->in_avail());

    if (avail == 0) {
        ++avail;
    } else if (buf_size < avail) {
        avail = buf_size;
    }
    is.read(buf, avail);

    std::streamsize count = is.gcount();

    if (count && is.eof() && !is.bad())
        is.clear();
    return count;
}

// PushbackStreamBuf definition ---------------------------------------------

const std::streamsize PushbackStreamBuf::k_MinBufSize = 4096;

PushbackStreamBuf::PushbackStreamBuf(
    std::istream & is,
    std::char_traits<char>::char_type* buf,
    std::streamsize    buf_size,
    void*         del_ptr) :
    m_Is(is),
    m_Buf(buf),
    m_BufSize(buf_size),
    m_DelPtr(del_ptr)
{
    setp(0, 0); //!< unbuffered output at this level of streambuf's hierarchy
    setg(m_Buf, m_Buf, m_Buf + m_BufSize);
    m_Sb = m_Is.rdbuf(this);
    
#ifdef HAVE_GOOD_IOS_CALLBACKS
    try {
        m_Index             = m_Is.xalloc();
        m_Is.pword(m_Index) = this;
        m_Is.register_callback(x_Callback, m_Index);
    }
    STD_CATCH_ALL("PushbackStreamBuf::PushbackStreamBuf");
#endif //!<HAVE_GOOD_IOS_CALLBACKS
    
}

// --------------------------------------------------------------------------

PushbackStreamBuf::~PushbackStreamBuf()
{
#ifdef HAVE_GOOD_IOS_CALLBACKS
    m_Is.pword(m_Index) = 0;
#endif //!<HAVE_GOOD_IOS_CALLBACKS

    delete[] (std::char_traits<char>::char_type*) m_DelPtr;
    if (m_Sb) {
        m_Is.rdbuf(m_Sb);
        if (dynamic_cast<PushbackStreamBuf*> (m_Sb))
            delete m_Sb;
    }
}

// --------------------------------------------------------------------------

std::char_traits<char>::pos_type PushbackStreamBuf::seekoff(
    std::char_traits<char>::off_type off,
    std::ios_base::seekdir whence,
    std::ios_base::openmode which)
{
    x_DropBuffer();
    if (whence == std::ios::cur  &&  (which & std::ios::in)) {
        return (std::char_traits<char>::pos_type)((std::char_traits<char>::off_type)(-1));
    }
    return m_Sb->pubseekoff(off, whence, which);
}
 
// --------------------------------------------------------------------------
 
std::char_traits<char>::pos_type PushbackStreamBuf::seekpos(
    std::char_traits<char>::pos_type pos,
    std::ios_base::openmode which)
{
    x_DropBuffer();
    return m_Sb->pubseekpos(pos, which);
}

// --------------------------------------------------------------------------

std::char_traits<char>::int_type PushbackStreamBuf::overflow(std::char_traits<char>::int_type c)
{
    if ( !std::char_traits<char>::eq_int_type(c, std::char_traits<char>::eof()) ) {
        return m_Sb->sputc(std::char_traits<char>::to_char_type(c));
    }

    return m_Sb->pubsync() == 0 ? std::char_traits<char>::not_eof(std::char_traits<char>::eof()) : std::char_traits<char>::eof();
}

// --------------------------------------------------------------------------

std::streamsize PushbackStreamBuf::xsputn(const std::char_traits<char>::char_type* buf, std::streamsize n)
{
    /// hope that this is an optimized copy operation (instead of overflow()s)
    return m_Sb->sputn(buf, n);
}

// --------------------------------------------------------------------------

std::char_traits<char>::int_type PushbackStreamBuf::underflow(void)
{
    /// we are here because there is no more data in the pushback buffer
    // underflow assert
    assert(gptr()  &&  gptr() >= egptr());

    x_FillBuffer(m_Sb->in_avail());
    return gptr() < egptr() ? std::char_traits<char>::to_int_type(*gptr()) : std::char_traits<char>::eof();
}

// --------------------------------------------------------------------------

std::streamsize PushbackStreamBuf::xsgetn(std::char_traits<char>::char_type* buf, std::streamsize m)
{
    size_t n_total = 0;
    while (m) {
        if (gptr() < egptr()) {
            size_t n       = (size_t) m;
            size_t n_avail = (size_t) (egptr() - gptr());
            size_t n_read  = (n <= n_avail) ? n : n_avail;
            memcpy(buf, gptr(), n_read*sizeof(std::char_traits<char>::char_type));
            gbump((int) n_read);
            m       -= (std::streamsize) n_read;
            buf     += (std::streamsize) n_read;
            n_total += (std::streamsize) n_read;
        } else {
            x_FillBuffer(m);
            if (gptr() >= egptr())
                break;
        }
    }
    return n_total;
}

// --------------------------------------------------------------------------

std::streamsize PushbackStreamBuf::showmanyc(void)
{
    /// we are here because (according to the standard) gptr() >= egptr()
    assert(gptr()  &&  gptr() >= egptr());

    return m_Sb->in_avail();
}

// --------------------------------------------------------------------------

std::char_traits<char>::int_type PushbackStreamBuf::pbackfail(std::char_traits<char>::int_type /*c*/)
{
    /* We always maintain "usual backup condition" (27.5.2.4.3.13) after
     * underflow(), i.e. 1 byte backup after a good read is always possible.
     * That is, this function gets called only if the user tries to
     * back up more than once (although, some attempts may be successful,
     * this function call denotes that the backup area is full).
     */
    return std::char_traits<char>::eof(); //!< always fail
}

// --------------------------------------------------------------------------

int PushbackStreamBuf::sync(void)
{
    return m_Sb->pubsync();
}

// --------------------------------------------------------------------------

std::streambuf * PushbackStreamBuf::setbuf(
    std::char_traits<char>::char_type* /*buf*/,
    std::streamsize    /*buf_size*/)
{
    throw "PushbackStreamBuf::setbuf: not allowed";
    return this;
}

// --------------------------------------------------------------------------

void PushbackStreamBuf::x_FillBuffer(std::streamsize max_size)
{
    assert(m_Sb);

    if ( !max_size )
        ++max_size;
        
    PushbackStreamBuf* sb = dynamic_cast<PushbackStreamBuf*> (m_Sb);
    if ( sb ) {
        assert(&m_Is == &sb->m_Is);
        m_Sb     = sb->m_Sb;
        sb->m_Sb = 0;
        if (sb->gptr() >= sb->egptr()) {
            delete sb;
            x_FillBuffer(max_size);
            return;
        }
        delete[] (std::char_traits<char>::char_type*) m_DelPtr;
        m_Buf        = sb->m_Buf;
        m_BufSize    = sb->m_BufSize;
        m_DelPtr     = sb->m_DelPtr;
        sb->m_DelPtr = 0;
        setg(sb->gptr(), sb->gptr(), sb->egptr());
        delete sb;
    } else {
        std::char_traits<char>::char_type* bp = 0;
        std::streamsize buf_size = m_DelPtr
            ? (std::streamsize)(m_Buf - (std::char_traits<char>::char_type*) m_DelPtr) + m_BufSize
            : 0;
        if (buf_size < k_MinBufSize) {
            buf_size = k_MinBufSize;
            bp = new std::char_traits<char>::char_type[buf_size];
        }
        std::streamsize n = m_Sb->sgetn(bp ? bp : (std::char_traits<char>::char_type*) m_DelPtr,
                                std::min(buf_size, max_size));
         if (n <= 0) {
            /// NB: For unknown reasons WorkShop6 can return -1 from sgetn :-/
            delete bp;
            return;
        }
        if (bp) {
            delete[] (std::char_traits<char>::char_type*) m_DelPtr;
            m_DelPtr = bp;
        }
        m_Buf = (std::char_traits<char>::char_type*) m_DelPtr;
        m_BufSize = buf_size;
        setg(m_Buf, m_Buf, m_Buf + n);
    }
}

// --------------------------------------------------------------------------

void PushbackStreamBuf::x_DropBuffer()
{
    PushbackStreamBuf* sb = dynamic_cast<PushbackStreamBuf*> (m_Sb);
    if (sb) {
        m_Sb     = sb->m_Sb;
        sb->m_Sb = 0;
        delete sb;
        x_DropBuffer();
        return;
    }
    /// nothing in the buffer; no putback area as well
    setg(m_Buf, m_Buf, m_Buf);
}

// --------------------------------------------------------------------------

} // namespace swift


