// swiftdatabasebuilder.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#include "swiftdatabasebuilder.hpp"
#include "binary_ostream_iterator.hpp"
#include "idmap.hpp"
#include "sequence.hpp"
#include "swiftdatabasebase.hpp"
#include <fstream>

namespace swift {

namespace detail {

// SwiftDatabaseBuilderImpl class declaration -------------------------------

struct SwiftDatabaseBuilderImpl : private SwiftDatabaseBase
{
    ~SwiftDatabaseBuilderImpl();

    void clear();
    void initialize();

    bool isOpen() const;
    void open(std::string const & name);
    void close();

    void writeIndex();

    std::string const & name() const;
    void setName(std::string const & name);

    void addSequence(
        Sequence const & sequence,
        std::string const & text,
        std::string const & id
    );


    std::string name_;
    IndexHeader header_;
    IdMapFactory::Sink idMap_;  
    OffsetVector textOffsets_;
    OffsetVector sequenceOffsets_;

    std::ofstream indexStream_;
    std::ofstream textStream_;
    std::ofstream sequenceStream_;
};

// SwiftDatabaseBuilderImpl class definition --------------------------------

/*!
    \internal

    Destroys the SwiftDatabaseBuilderImpl object.
*/
SwiftDatabaseBuilderImpl::~SwiftDatabaseBuilderImpl()
{
    close();
    clear();
}


/*!
    \internal

    Clears data structures for database. 
*/
void SwiftDatabaseBuilderImpl::clear()
{
    idMap_.clear();
    OffsetVector().swap(textOffsets_);
    OffsetVector().swap(sequenceOffsets_);
}


/*!
    \internal

    Initializes data structures for database. 
*/
void SwiftDatabaseBuilderImpl::initialize()
{
    textOffsets_.push_back(0);
    sequenceOffsets_.push_back(0);
}


/*!
    \internal

    Returns true if output files have been opened.
*/
bool SwiftDatabaseBuilderImpl::isOpen() const
{
    return !name_.empty();
}


/*!
    \internal

    Opens the output files using \a name as basename. Afterwards,
    name() returns \a name.

    \sa name()
*/
void SwiftDatabaseBuilderImpl::open(std::string const & name)
{
    indexStream_.open((name + IndexSuffix).c_str(), std::ios::binary);
    textStream_.open((name + TextSuffix).c_str(), std::ios::binary);
    sequenceStream_.open((name + SequenceSuffix).c_str(), std::ios::binary);

    name_ = name;
}


/*!
    \internal

    Writes remaining data and closes all output files. Afterwards, name()
    returns the empty string.    

    \sa name()
*/
void SwiftDatabaseBuilderImpl::close()
{
    if (!isOpen())
        return;

    writeIndex();

    indexStream_.close();
    textStream_.close();
    sequenceStream_.close();

    name_.clear();
}


std::string const & SwiftDatabaseBuilderImpl::name() const
{
    return name_;
}


void SwiftDatabaseBuilderImpl::setName(std::string const & name)
{
    if (isOpen()) {
        close();
        clear();
    }

    open(name);
    initialize();
}


/*!
    \internal 

    Adds the sequence \a sequence with the text \a text to the database.
    If the id \a id is non-empty, the sequence is additionally registered
    under the id.
*/
void SwiftDatabaseBuilderImpl::addSequence(
    Sequence const & sequence,
    std::string const & text,
    std::string const & id
)
{
    if (!isOpen())
        return;

    if (!id.empty()) {
        idMap_.write(id, sequenceOffsets_.size());
    }

    sequenceStream_.write(
        reinterpret_cast<char const *>(sequence.const_data()),
        sequence.size()
    );
    sequenceOffsets_.push_back(sequenceOffsets_.back() + sequence.size());

    textStream_ << text << '\0';
    textOffsets_.push_back(textOffsets_.back() + text.size() + 1);
}


/*!
    \internal

    Writes header and indexes to index stream.
*/
void SwiftDatabaseBuilderImpl::writeIndex()
{
    header_.constructionDate = time(0);
    header_.sequenceCount = sequenceOffsets_.size() - 1;

    // Write header
    indexStream_.write(
        reinterpret_cast<char const *>(&header_),
        sizeof(IndexHeader)
    );

    // Write text offsets
    std::copy(
        textOffsets_.begin(),
        textOffsets_.end(),
        binary_ostream_iterator<OffsetVector::value_type>(indexStream_)
    );

    // Write sequence offsets
    std::copy(
        sequenceOffsets_.begin(),
        sequenceOffsets_.end(),
        binary_ostream_iterator<OffsetVector::value_type>(indexStream_)
    );

    // Write id map
    idMap_.write(indexStream_);
}


} // namespace detail

// SwiftDatabaseBuilder class definition ------------------------------------

/*!
    Constructs an empty SwiftDatabaseBuilder object. Before adding sequences,
    you must call setName().

    \sa setName()
*/
SwiftDatabaseBuilder::SwiftDatabaseBuilder() :
    pimpl_(boost::shared_ptr<ImplType>(new ImplType))
{
}


/*!
    Destroys the SwiftDatabaseBuilder object.
*/
SwiftDatabaseBuilder::~SwiftDatabaseBuilder()
{
}


std::string const & SwiftDatabaseBuilder::name() const
{
    return pimpl_->name();
}


void SwiftDatabaseBuilder::setName(std::string const & name)
{
    pimpl_->setName(name);
}


void SwiftDatabaseBuilder::addSequence(
    Sequence const & sequence,
    std::string const & text,
    std::string const & id
)
{
    pimpl_->addSequence(sequence, text, id); 
}

// --------------------------------------------------------------------------


} // namespace swift

