// timer.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#ifndef _SWIFT_TIMER_
#define _SWIFT_TIMER_

extern "C"
{
#include <time.h>
#include <sys/types.h>
#include <sys/resource.h>
}

#include <iostream>
#include <string>

namespace swift {


class TimeElapsed
{
    public:
    TimeElapsed(void);
    TimeElapsed(long userTime, long systemTime, long realTime);
    TimeElapsed(TimeElapsed const & timeElapsed);
    ~TimeElapsed();

    TimeElapsed& operator=(const TimeElapsed& source);
    const TimeElapsed& operator+=(const TimeElapsed& source);
    const TimeElapsed& operator-=(const TimeElapsed& source);

    long getUserTime() const;
    double getUserTimeSeconds() const;
    long getSystemTime() const;
    double getSystemTimeSeconds() const;
    long getRealTime() const;
    double getRealTimeSeconds() const;

    private:
    long userTime_;
    long systemTime_;
    long realTime_;

    friend TimeElapsed operator+(const TimeElapsed& lhs,
                                 const TimeElapsed& rhs);
    friend TimeElapsed operator-(const TimeElapsed& lhs,
                                 const TimeElapsed& rhs);

    friend std::ostream & operator<<(
        std::ostream & outStream, 
        TimeElapsed& timeElapsed
    );
};

TimeElapsed operator+(const TimeElapsed& lhs, const TimeElapsed& rhs);
TimeElapsed operator-(const TimeElapsed& lhs, const TimeElapsed& rhs);
std::ostream & operator<<(std::ostream & outStream, TimeElapsed& timeElapsed);

std::string timeToString(long time);


/*!
    Timer class

    Note: When using the usec precision the timer will rollover every
          72 minutes due to the limited precision (32 bit) of the system
          libraries.
*/
class Timer
{
    public:
    Timer(bool useUSecPrecision = true);
    virtual ~Timer();

    void start();
    void stop();
    void reset();
    TimeElapsed getTimeElapsed();
    bool isRunning();

    private:
    bool isRunning_;
    bool useUSecPrecision_;

    time_t timeBegin_;
    time_t timeEnd_;

    struct rusage usageBegin_;
    struct rusage usageEnd_;

    long userTime_;
    long systemTime_;
    long realTime_;
};


inline
long TimeElapsed::getUserTime() const
{
    return userTime_;
}


inline
double TimeElapsed::getUserTimeSeconds() const
{
    return userTime_ / 1000000.0;
}


inline
long TimeElapsed::getSystemTime() const
{
    return systemTime_;
}


inline
double TimeElapsed::getSystemTimeSeconds() const
{
    return systemTime_ / 1000000.0;
}


inline
long TimeElapsed::getRealTime() const
{
    return realTime_;
}


inline 
double TimeElapsed::getRealTimeSeconds() const
{
    return realTime_ / 1000000.0;
}


} // namespace swift

#endif


