// unique_keep.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#ifndef _SWIFT_UNIQUE_KEEP_
#define _SWIFT_UNIQUE_KEEP_

namespace swift {

// --------------------------------------------------------------------------

/*!
    Variation of std::unique<>() which does not trash the duplicate elements.
    
    unique_keep takes a sorted range, and returns an iterator   newLast  such
    that the range [first, newLast) contains no two consecutive elements
    that are duplicates with respect to the predicate \a binary_pred.

    On return, duplicate elements are found in the range [newLast, last).
*/

template <
    typename ForwardIterator,
    typename BinaryPredicate
>
ForwardIterator unique_keep(
    ForwardIterator first,
    ForwardIterator last,
    BinaryPredicate binary_pred
)
{
    if (first == last) 
        return first; 

    ForwardIterator prev(first);
    while (++first != last)
        if (!binary_pred(*first, *prev))
            std::swap(*++prev,*first);
    return ++prev;
}

// --------------------------------------------------------------------------

} // namespace swift

#endif

