// epsilonaligner.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#include "epsilonaligner.hpp"
#include "karlinaltschul.hpp"
#include "scorematrix.hpp"
#include "sequencedatabase.hpp"

namespace swift {


/*!
    Constructs an EpsilonAligner object.
*/
EpsilonAligner::EpsilonAligner()
{
}


/*!
    Destroys the EpsilonAligner object.
*/
EpsilonAligner::~EpsilonAligner()
{
}


void EpsilonAligner::setGramParameters(
    unsigned int gramSize,
    unsigned int maxGap,
    unsigned int minHits
)
{
    gcaParams_.maxGap   = maxGap;
    gcaParams_.minHits  = minHits;
    gcaParams_.gramSize = gramSize;
}
 

void EpsilonAligner::setSeedingScoreFunction(
    ScoreFunction const & scoreFunction,
    int matchScore
)
{
    gcaAligner_.setScoreFunction(scoreFunction);
    gcaAligner_.setMatchScore(matchScore);
}


void EpsilonAligner::setSeedingMaskSensitivity(MaskSensitivity ms)
{
    gcaAligner_.setQueryMaskSensitivity(ms);
}


void EpsilonAligner::setExtensionScoreFunction(
    ScoreFunction const & scoreFunction,
    int xDrop
)
{
    semiAligner_.setScoreFunction(scoreFunction);
    semiAligner_.setDropOff(xDrop);
}


void EpsilonAligner::findGramChainAlignments(
    Sequence const & query,
    Sequence const & subject,
    EpsilonRegion const & region
)
{
    ASSERT(region.queryEnd <= query.size(), "region query end out of bounds"); 

    unsigned int subjectStart(std::max(0, region.subjectStart()));
    unsigned int subjectEnd(std::min(region.subjectEnd(), (unsigned int)subject.size()));
    int diagonalOffset(region.queryStart - subjectStart);

    gcaParams_.query         = 
        reinterpret_cast<char const *>(query.const_data() + region.queryStart);
    gcaParams_.queryLength   = region.querySize();

    gcaParams_.subject       =
        reinterpret_cast<char const *>(subject.const_data() + subjectStart);
    gcaParams_.subjectLength = subjectEnd - subjectStart;

    gcaParams_.lowerDiagonal   = region.minDiagonal + diagonalOffset;
    gcaParams_.upperDiagonal   = region.maxDiagonal + diagonalOffset;

    gcaAligner_.alignments(gcaParams_, gcaResult_);

    queryOffset_   = region.queryStart;
    subjectOffset_ = subjectStart;
}


} // end namespace swift


