// epsilonaligner.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#ifndef _SWIFT_EPSILONALIGNER_
#define _SWIFT_EPSILONALIGNER_

#include "alignment.hpp"
#include "assert.hpp"
#include "scorefunction.hpp"
#include "masksensitivity.hpp"
#if 0
#include "gappedxdropaligner.hpp"
#endif
#include "semigappedxdropaligner.hpp"
#include "gramchainaligner.hpp"
#include "sequence.hpp"
#include <vector>


namespace swift {

struct EpsilonRegion;


class EpsilonAligner
{
    public:
    EpsilonAligner();
    ~EpsilonAligner();

    void setSeedingScoreFunction(
        ScoreFunction const & scoreFunction,
        int matchScore
    );

    void setSeedingMaskSensitivity(MaskSensitivity ms);

    void setExtensionScoreFunction(
        ScoreFunction const & scoreFunction,
        int xDrop
    );

    void setGramParameters(
        unsigned int gramSize,
        unsigned int maxGap,
        unsigned int minHits
    );

    template <class OutputIterator>
    void alignments(
        Sequence const & query,
        Sequence const & subject,
        EpsilonRegion const & region,
        OutputIterator output
    );

    private:
    void findGramChainAlignments(
        Sequence const & query,
        Sequence const & subject,
        EpsilonRegion const & region
    );

    GramChainAligner             gcaAligner_;
    GramChainAligner::Parameters gcaParams_;
    GramChainAligner::Result     gcaResult_;
    
#if 0
    GappedXDropAligner             gxpAligner_;
    GappedXDropAligner::Parameters gxpParams_;
    GappedXDropAligner::Result     gxpResult_;
#endif

    SemiGappedXDropAligner             semiAligner_;
    SemiGappedXDropAligner::Parameters semiParams_;
    SemiGappedXDropAligner::Result     semiResult_;


    unsigned int queryOffset_;
    unsigned int subjectOffset_;
};


/*!
    Finds epsilon alignments between \a query and \a subject, that pass
    through the epsilon region \a region. The alignments are output to
    the output iterator \a output, which must have value type
    <tt>Alignments *</tt>.
*/
template <class OutputIterator>
void EpsilonAligner::alignments(
    Sequence const & query,
    Sequence const & subject,
    EpsilonRegion const & region,
    OutputIterator output
)
{
    using std::vector;

    findGramChainAlignments(query, subject, region);

#if 0
    gxpParams_.query         = query.const_data();
    gxpParams_.queryLength   = query.size();
    gxpParams_.subject       = subject.const_data();
    gxpParams_.subjectLength = subject.size();
#endif
    semiParams_.query       = query.const_data();
    semiParams_.querySize   = query.size();
    semiParams_.subject     = subject.const_data();
    semiParams_.subjectSize = subject.size();

    typedef vector<Alignment *>::iterator MyIterator;
    for (MyIterator i(gcaResult_.begin()), j(gcaResult_.end()); i != j; ++i) {
        Alignment * a(*i);

        // Align to the left
#if 0
        gxpParams_.queryBegin     = queryOffset_ + a->queryBegin;
        gxpParams_.subjectBegin   = subjectOffset_ + a->subjectBegin;
        gxpParams_.alignDirection = GappedXDropAligner::AlignLeft;
        gxpParams_.ungappedLength = 0;

        gxpAligner_.getAlignment(gxpParams_, gxpResult_);
#endif
        semiParams_.queryOffset = queryOffset_ + a->queryBegin;
        semiParams_.subjectOffset = subjectOffset_ + a->subjectBegin;
        semiAligner_.findAlignmentBeforeSeed(semiParams_, semiResult_);

        // Set alignment start point
#if 0
        a->queryBegin      = gxpResult_.queryBegin;
        a->subjectBegin    = gxpResult_.subjectBegin;
        a->score          += gxpResult_.score;
#endif
        a->queryBegin      = semiResult_.queryBegin;
        a->subjectBegin    = semiResult_.subjectBegin;
        a->score          += semiResult_.score;
                            
        // Align to the right
#if 0
        gxpParams_.queryBegin     = queryOffset_ + a->queryEnd;
        gxpParams_.subjectBegin   = subjectOffset_ + a->subjectEnd;
        gxpParams_.alignDirection = GappedXDropAligner::AlignRight;
        gxpParams_.ungappedLength = 0;

        gxpAligner_.getAlignment(gxpParams_, gxpResult_);
#endif
        semiParams_.queryOffset = queryOffset_ + a->queryEnd;
        semiParams_.subjectOffset = subjectOffset_ + a->subjectEnd;
        semiAligner_.findAlignmentAfterSeed(semiParams_, semiResult_);
           
        // Set alignment end point
#if 0
        a->queryEnd        = gxpResult_.queryEnd;
        a->subjectEnd      = gxpResult_.subjectEnd;
        a->score          += gxpResult_.score;
#endif
        a->queryEnd    = semiResult_.queryEnd;
        a->subjectEnd  = semiResult_.subjectEnd;
        a->score      += semiResult_.score;

        *output = a;
        ++output;
    }

    gcaResult_.clear();
}


/*!
    Defines an epsilon region in an implied dynamic programming matrix.
*/
struct EpsilonRegion
{
    EpsilonRegion(
        unsigned int qs,
        unsigned int qe,
        int mind,
        int maxd
    );

    unsigned int querySize() const;
    int subjectStart() const;
    unsigned int subjectEnd() const;

    unsigned int queryStart;
    unsigned int queryEnd;
    int minDiagonal;
    int maxDiagonal;
};


/*!
    Constructs an epsilon region.

    The half-open range <tt>[qs, qe)</tt> defines the substring of the
    query sequence that is coverered. The closed range
    <tt>[mind, maxd]</tt> defines the covered diagonals.
*/
inline
EpsilonRegion::EpsilonRegion(
    unsigned int qs,
    unsigned int qe,
    int mind,
    int maxd
) :
    queryStart(qs),
    queryEnd(qe),
    minDiagonal(mind),
    maxDiagonal(maxd)
{
}


/*!
    Returns the size of the epsilon region when projected onto the
    query sequence.
*/
inline
unsigned int EpsilonRegion::querySize() const
{
    return queryEnd - queryStart;
}


/*!
    Returns \c subjectStart from the half-open range
    <tt>[subjectStart, subjectEnd)</tt> which defines the substring
    of the subject sequence that is covered by the epsilon region.
    The returned position can be negative.
*/
inline
int EpsilonRegion::subjectStart() const
{
    return queryStart + minDiagonal;
}


/*!
    Returns \c subjectEnd from the half-open range
    <tt>[subjectStart, subjectEnd)</tt> which defines the substring
    of the subject sequence that is covered by the epsilon region.
    The returned position may be larger than the size of the
    subject sequence.
*/
inline
unsigned int EpsilonRegion::subjectEnd() const
{
    return queryEnd + maxDiagonal;
}


} // namespace swift

#endif

