// epsilonfilter.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#include "epsilonfilter.hpp"
#include "epsilonzone.hpp"
#include <iostream>
#include <cstring>


namespace swift {

/*!
    \struct EpsilonFilter::Match

    Defines a parallelogramic region of the implied dynamic programming
    that may contain an epsilon match. The parallelogram covers these
    half-open ranges in the query and subject, respectively:

    \li <tt>[queryStart, queryEnd)</tt>
    \li <tt>[subjectStart, subjectEnd)</tt>

    The coordinates are obtained by projecting the parallelogram onto
    the query resp. the subject. Therefore, \c subjectStart may be negative
    and \c subjectEnd may be larger than the subject size.
*/


/*!
    Constructs an epsilon filter using the gram index \a gramIndex.
*/
EpsilonFilter::EpsilonFilter(GramIndex const & gramIndex) :
    taggedBins_(32768, 0.75),
    gramIndex_(gramIndex),
    subjectSize_(gramIndex.maxPosition() + gramIndex.gramSize()),
    gramSize_(gramIndex.gramSize()),
    bins_(0)
{
}


/*!
    Destroys the EpsilonFilter object.
*/
EpsilonFilter::~EpsilonFilter()
{
    delete [] bins_;
}


bool EpsilonFilter::configure(
    unsigned int minLength,
    double epsilon,
    unsigned int minBinShift
) throw()
{
    EpsilonZone zone(minLength, gramSize_, epsilon);
    if (!zone.isValid())
        return false;

    // Determine a bin shift
    binShift_ = 0;
    for (unsigned int b = 1, n = sizeof(unsigned int) * 8 - 1; binShift_ < n; ++binShift_) {
        if (b > zone.width())
            break;
        b <<= 1;
    }
    if (binShift_ < minBinShift)
        binShift_ = minBinShift;

    binMask_   = (1UL << binShift_) - 1;
    binSize_   = (1UL << binShift_) + zone.width();
    zoneWidth_ = zone.width();
    threshold_ = zone.threshold();
    trapMin_   = zone.height() - gramSize_;
    minMatchLength_ = zone.threshold() + gramSize_ - 1;

    binMax_ = ((subjectSize_ + binSize_ - 1) >> binShift_) + 1;
    try {
        delete [] bins_;
        bins_ = new Bin[binMax_];
    } catch (std::bad_alloc & e) {
            std::cerr << "Memory allocation failed in epsilon filter. "
                     "Aborting." << std::endl;
        exit(1);
    }
    memset(bins_, 0, sizeof(Bin) * binMax_);

    for (unsigned int i(0); i < binMax_; ++i)
        bins_[i].last = -1; 
          
    ic_ = 0;
    id_ = -trapMin_;
    
    matchProperties_.width   = binSize_;
    // FIXME 1148737215
    // matchProperties_.maxGap  = zone.width - 1;  // CHANGED!
    matchProperties_.maxGap  = zone.width();
    matchProperties_.minHits = zone.threshold();
    matchProperties_.gramSize = gramSize_;
    
    return true;
}


std::ostream & operator<<(
    std::ostream & out,
    EpsilonFilter::MatchProperties const & mp
)
{
    out <<
        "MatchProperties (width = " << mp.width <<
        ", maxgap = " << mp.maxGap << 
        ", minhits = " << mp.minHits << ")";
    return out;
}


} // namespace swift

