// indexcommand.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//
//
// Additions / changes by Peter Husemann
//

#include "indexcommand.hpp"
#include "gramindex.hpp"
#include "swiftdatabase.hpp"
#include <boost/scoped_ptr.hpp>
#include <algorithm>
#include <iostream>

namespace swift
{

/*!
    Constructs a IndexCommand object.
*/
IndexCommand::IndexCommand() :
    visibleOptions_("Valid options")
{
    namespace po = boost::program_options;

    visibleOptions_.add_options()
        ("help,h", "produces this help message.")
        ("verbose,v", "print extra information.")
        ("mask", po::bool_switch(),
         "do not index q-grams in masked sequence. "
         "Soft masking (lower-case letters) and hard masking "
         "(Ns) is respected.");

    hiddenOptions_.add_options()
        ("database", po::value<std::string>(&databasePrefix_),
        "Database prefix")
        ("gram", po::value<std::vector<unsigned int> >(&gramSizes_),
        "Index gram size");
}


void IndexCommand::execute(int argc, char ** argv)
{
  //SwiftCommand::SwiftExecutable = argv[0];
  if (processArguments(argc, argv))
        indexDatabase();
}


bool IndexCommand::processArguments(int argc, char ** argv)
{
    namespace po = boost::program_options;

    gramSizes_.clear();

    try {
        po::positional_options_description positionals;
        positionals.add("database", 1);
        positionals.add("gram", -1);

        po::options_description all;
        all.add(visibleOptions_).add(hiddenOptions_);

        po::variables_map vm;
        po::store(po::command_line_parser(argc, argv).
                  options(all).positional(positionals).run(), vm);
        po::notify(vm);

        if (vm.count("help") || vm.count("gram") < 1) {
            displayUsage();
            return false;
        }

        maskSensitivity_ = vm["mask"].as<bool>() ?
                           MaskSensitive : MaskInsensitive;

    } catch (po::unknown_option & e) {
      std::cerr << SwiftCommand::SwiftExecutable <<": " << e.what() << "\n"
            << "Type '"<< SwiftCommand::SwiftExecutable <<" help' for usage information."
                  << std::endl;
        return false;
    } catch (std::exception & e) {
        std::cerr << e.what() << std::endl;
        return false;
    }

    return true;
}

void IndexCommand::displayUsage(std::ostream & outStream) const
{
    outStream <<
        "Usage:\n"
        "  "<< SwiftCommand::SwiftExecutable <<" index [options] DATABASE [GRAM...]\n"
        "\n"
        "  Constructs a q-gram index for the SWIFT database DATABASE.\n"
        "  The argument GRAM specifies the size of the q-grams to index.\n"
        "  Multiple indexes can be constructed by issuing more than one\n"
        "  gram size.\n"
        "\n"
        "  The constructed index consists of two files with name\n"
        "  prefix DATABASE and the following suffices:\n"
        "\n"
        "    _GRAM.swl   -   index meta-information\n"
        "    _GRAM.swp   -   q-gram positions\n"
        "\n"
        "  Examples:\n"
        "    "<< SwiftCommand::SwiftExecutable <<" index --mask mydatabase 11\n"
        "\n"
        "\n"
        "  You have to create an database beforehand. See:\n"
        "    create\n"
        "\n"
        << visibleOptions_ << std::endl;
}


void IndexCommand::indexDatabase()
{
    SwiftDatabase database(databasePrefix_);

    if (!database.isOpen()) {
        std::cout << "Cannot open database with prefix '" << databasePrefix_
            << "'" << std::endl;
        return;
    }

    std::sort(gramSizes_.begin(), gramSizes_.end());
    for (std::vector<unsigned int>::const_iterator i(gramSizes_.begin()),
         j(std::unique(gramSizes_.begin(), gramSizes_.end())); i != j; ++i) {

        boost::scoped_ptr<GramIndex> index(
            GramIndex::construct(&database, *i, maskSensitivity_)
        );

        if (!index) {
            std::cout << "Cannot construct index for database with '"
                << databasePrefix_ << "', using gram size " << *i << std::endl;
            return;
        }

        index->serialize(databasePrefix_);
    }
}


} // namespace swift
