// versioncommand.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//
//
// Additions / changes by Peter Husemann
//

#include "versioncommand.hpp"
#include "version.hpp"
#include <iostream>

namespace swift
{


/*!
    Constructs a VersionCommand object.
*/
VersionCommand::VersionCommand() :
    visibleOptions_("Valid options")
{
    namespace po = boost::program_options;

    visibleOptions_.add_options()
        ("help,h", "produces this help message.")
        ("quiet,q", po::bool_switch(),
         "print as little as possible.");
}


void VersionCommand::execute(int argc, char ** argv)
{
  //SwiftCommand::SwiftExecutable = argv[0];
  if (processArguments(argc, argv))
        displayVersion();
}


bool VersionCommand::processArguments(int argc, char ** argv)
{
    namespace po = boost::program_options;

    try {
        po::options_description all;
        all.add(visibleOptions_);

        po::variables_map vm;
        po::store(po::command_line_parser(argc, argv).options(all).run(), vm);
        po::notify(vm);

        if (vm.count("help")) {
            displayUsage();
            return false;
        }

        quiet_ = vm["quiet"].as<bool>();

    } catch (po::unknown_option & e) {
      std::cerr << SwiftCommand::SwiftExecutable <<": " << e.what() << "\n"
                  << "Type 'swift help' for usage information."
                  << std::endl;
        return false;
    } catch (std::exception & e) {
        std::cerr << e.what() << std::endl;
        return false;
    }

    return true;
}


void VersionCommand::displayUsage(std::ostream & outStream) const
{
    outStream <<
        "Usage:\n"
        "  swift version [options]\n"
        "\n"
        "  Displays the version of the Swift exectuable and the.\n"
        "  copyright and licensing information.\n"
        "\n"
        << visibleOptions_ << std::endl;
}


void VersionCommand::displayVersion() const
{
    if (quiet_) {
        std::cout << Version::svnRevision() << "\n";
        return;
    }

    std::cout
        << "swift r" << Version::svnRevision() << "\n"
        << Version::copyright() << "\n\n"
        << Version::license() << std::endl;
}


} // namespace swift
