#!/bin/bash

WORK_DIR="`pwd`/validation_work"
REF_PATH="`pwd`/reference.fasta"
SCAF_PATH="`pwd`/scaffolds.fasta"
GRAM=11
MAXGAP=0
MAXERROR=0.05
MINLEN=30

usage() {
	echo validate_local.sh: A script for running validation on a single machine.
	echo
	echo "Usage:"
	echo "Parameter		(Default value)"
	echo
	echo "--------------------------------------------"
	echo "-h | --help			Display this text"
	echo "--maxerror VALUE	(=$MAXERROR)	Allow error rate of VALUE when doing alignments with swift"
	echo "--minlen VALUE		(=$MINLEN)	Only report alignments with length >= VALUE"
	echo "--maxgap VALUE		(=$MAXGAP)	Restrict gaps to at max VALUE when doing colinear chaining (0 = unlimited)"
	echo "--gram VALUE		(=$GRAM)	Use VALUE-gram for swift"
	echo "-r | --reference PATH	(=$REF_PATH)	Path to reference sequence"
	echo "-s | --scaffolds PATH	(=$SCAF_PATH)	Path to scaffolds"
	echo "-w | --work PATH	(=$WORK_DIR)	Where to store intermediary and final results"
	echo
}

SCRIPT_DIR="`dirname $0`"

if [ $# -lt 2 ]; then
	usage
	exit 1
fi

while [ $# -gt 0 ]; do
	case $1 in 
		-h|--help) usage; exit 1;;
		--maxerror) MAXERROR=$2; shift 2;;
		--minlen) MINLEN=$2; shift 2;;
		--maxgap) MAXGAP=$2; shift 2;;
		--gram) GRAM=$2; shift 2;;
		-r|--reference) REF_PATH=$2; shift 2;;
		-s|--scaffolds) SCAF_PATH=$2; shift 2;;
		-w|--work) WORK_DIR=$2; shift 2;;
		*) shift 1;;
	esac
done

SCAFID=`basename $SCAF_PATH`
SCAFID=${SCAFID%.*}
REFID="`basename $REF_PATH`"
REFID=${REFID%.*}

IDENTIFIER="${REFID}_$SCAFID"

echo "Using settings"
echo "--------------"
echo "Maxerror	$MAXERROR"
echo "Minlength	$MINLEN"
echo "Maxgap		$MAXGAP"
echo "Gram		$GRAM"
echo
echo "Work dir	$WORK_DIR"
echo "Reference	$REF_PATH"
echo "Scaffolds	$SCAF_PATH"
echo "Identifier	$IDENTIFIER"

if [ ! -f $SCAF_PATH ]; then
	echo "Could not find file $SCAF_PATH"
	exit 1
fi

if [ ! -f $REF_PATH ]; then
	echo "Could not find file $REF_PATH"
	exit 1
fi
echo "--------------"

## Create working directory if doesn't exist
mkdir -p $WORK_DIR
## Maintain compatibility with the rest of the pipeline
echo "cat $SCAF_PATH | $SCRIPT_DIR/bin/rename_sequences > $WORK_DIR/reordered_scafs.fasta"
cat $SCAF_PATH | $SCRIPT_DIR/bin/rename_sequences > $WORK_DIR/reordered_scafs.fasta

## Concatenate reference sequence
echo "$SCRIPT_DIR/bin/concatenate_fasta $REF_PATH $WORK_DIR/concat_reference_${REFID}.fasta $WORK_DIR/boundaries_$REFID"
$SCRIPT_DIR/bin/concatenate_fasta $REF_PATH $WORK_DIR/concat_reference_${REFID}.fasta $WORK_DIR/boundaries_$REFID

## Build swift index
echo "$SCRIPT_DIR/swift/build/swift/swift create $WORK_DIR/database_$REFID $WORK_DIR/concat_reference_${REFID}.fasta"
$SCRIPT_DIR/swift/build/swift/swift create $WORK_DIR/database_$REFID $WORK_DIR/concat_reference_${REFID}.fasta
echo "$SCRIPT_DIR/swift/build/swift/swift index $WORK_DIR/database_$REFID --gram $GRAM"
$SCRIPT_DIR/swift/build/swift/swift index $WORK_DIR/database_$REFID --gram $GRAM
echo "$SCRIPT_DIR/swift/build/swift/swift create $WORK_DIR/query_$SCAFID $WORK_DIR/reordered_scafs.fasta"
$SCRIPT_DIR/swift/build/swift/swift create $WORK_DIR/query_$SCAFID $WORK_DIR/reordered_scafs.fasta
echo "$SCRIPT_DIR/swift/build/swift/swift index $WORK_DIR/query_$SCAFID --gram $GRAM"
$SCRIPT_DIR/swift/build/swift/swift index $WORK_DIR/query_$SCAFID --gram $GRAM

## Execute swift queries
echo "$SCRIPT_DIR/swift/build/swift/swift query --maxerror $MAXERROR --minlen $MINLEN --gram $GRAM $WORK_DIR/database_$REFID $WORK_DIR/query_$SCAFID > $WORK_DIR/$IDENTIFIER.results" 
$SCRIPT_DIR/swift/build/swift/swift query --maxerror $MAXERROR --minlen $MINLEN --gram $GRAM $WORK_DIR/database_$REFID $WORK_DIR/query_$SCAFID > $WORK_DIR/$IDENTIFIER.results

## Get scaffold lengths
echo "$SCRIPT_DIR/bin/calc_estl $WORK_DIR/reordered_scafs.fasta > $WORK_DIR/estl_$SCAFID"
$SCRIPT_DIR/bin/calc_estl $WORK_DIR/reordered_scafs.fasta > $WORK_DIR/estl_$SCAFID

## Extract tuples from swift results
echo "$SCRIPT_DIR/bin/extract_tuples $WORK_DIR/$IDENTIFIER.results $WORK_DIR/estl_$SCAFID $WORK_DIR/tuples_$IDENTIFIER"
$SCRIPT_DIR/bin/extract_tuples $WORK_DIR/$IDENTIFIER.results $WORK_DIR/estl_$SCAFID $WORK_DIR/tuples_$IDENTIFIER

## Perform colinear chaining
echo "$SCRIPT_DIR/bin/col_chain $WORK_DIR/tuples_$IDENTIFIER $WORK_DIR/boundaries_$REFID $WORK_DIR/correct_alignments_$IDENTIFIER $MAXGAP"
$SCRIPT_DIR/bin/col_chain $WORK_DIR/tuples_$IDENTIFIER $WORK_DIR/boundaries_$REFID $WORK_DIR/correct_alignments_$IDENTIFIER $MAXGAP

GENOME_LENGTH=`grep '^[^>]' $WORK_DIR/concat_reference_${REFID}.fasta | tr -d '\n' | wc -m`
NUMSCAFFOLDS=`grep '^>' $WORK_DIR/reordered_scafs.fasta | wc -l`

## Calculate final values
echo "$SCRIPT_DIR/bin/calc_coverages $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/reordered_scafs.fasta $GENOME_LENGTH > $WORK_DIR/validation_results.txt"
$SCRIPT_DIR/bin/calc_coverages $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/reordered_scafs.fasta $GENOME_LENGTH > $WORK_DIR/validation_results.txt
echo "$SCRIPT_DIR/bin/calc_norm_n50 $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/reordered_scafs.fasta >> $WORK_DIR/validation_results.txt"
$SCRIPT_DIR/bin/calc_norm_n50 $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/reordered_scafs.fasta $GENOME_LENGTH >> $WORK_DIR/validation_results.txt
echo "$SCRIPT_DIR/src/png_script.sh $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/${IDENTIFIER} $NUMSCAFFOLDS"
$SCRIPT_DIR/src/png_script.sh $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/${IDENTIFIER} $NUMSCAFFOLDS
cat $WORK_DIR/validation_results.txt
